import os
import tempfile

import pytest

from .... import examples as pysal_examples
from ...fileio import FileIO
from ..arcgis_swm import ArcGISSwmIO


class TesttestArcGISSwmIO:
    def setup_method(self):
        self.test_file = test_file = pysal_examples.get_path("ohio.swm")
        self.obj = ArcGISSwmIO(test_file, "r")

    def test_close(self):
        f = self.obj
        f.close()
        pytest.raises(ValueError, f.read)

    def test_read(self):
        w = self.obj.read()
        assert w.n == 88
        assert w.mean_neighbors == 5.25
        assert list(w[1].values()) == [1.0, 1.0, 1.0, 1.0]

    def test_seek(self):
        self.test_read()
        pytest.raises(StopIteration, self.obj.read)
        self.obj.seek(0)
        self.test_read()

    def test_write(self):
        w = self.obj.read()
        f = tempfile.NamedTemporaryFile(suffix=".swm")
        fname = f.name
        f.close()
        o = FileIO(fname, "w")
        o.write(w)
        o.close()
        wnew = FileIO(fname, "r").read()
        assert wnew.pct_nonzero == w.pct_nonzero
        os.remove(fname)



"""
Computational geometry code for PySAL: Python Spatial Analysis Library.

"""

__author__ = "Sergio J. Rey, Xinyue Ye, Charles Schmidt, Andrew Winslow, Hu Shao"

# ruff: noqa: A003, N802, E402

import math
import warnings

from .sphere import arcdist

__all__ = [
    "Point",
    "LineSegment",
    "Line",
    "Ray",
    "Chain",
    "Polygon",
    "Rectangle",
    "asShape",
]


dep_msg = (
    "Objects based on the `Geometry` class will deprecated "
    "and removed in a future version of libpysal."
)


def asShape(obj):
    """Returns a PySAL shape object from ``obj``, which
    must support the ``__geo_interface__``.

    Parameters
    ----------
    obj : {libpysal.cg.{Point, LineSegment, Line, Ray, Chain, Polygon}
        A geometric representation of an object.

    Raises
    ------
    TypeError
        Raised when ``obj`` is not a supported shape.
    NotImplementedError
        Raised when ``geo_type`` is not a supported type.

    Returns
    -------
    obj : {libpysal.cg.{Point, LineSegment, Line, Ray, Chain, Polygon}
        A new geometric representation of the object.
    """

    if isinstance(obj, Point | LineSegment | Line | Ray | Chain | Polygon):
        pass
    else:
        geo = obj.__geo_interface__ if hasattr(obj, "__geo_interface__") else obj

        if hasattr(geo, "type"):
            raise TypeError(f"{obj!r} does not appear to be a shape object.")

        geo_type = geo["type"].lower()

        # if geo_type.startswith('multi'):
        #    raise NotImplementedError, "%s are not supported at this time."%geo_type

        if geo_type in _geoJSON_type_to_Pysal_type:
            obj = _geoJSON_type_to_Pysal_type[geo_type].__from_geo_interface__(geo)
        else:
            raise NotImplementedError(f"{geo_type} is not supported at this time.")

    return obj


class Geometry:
    """A base class to help implement ``is_geometry``
    and make geometric types extendable.
    """

    def __init__(self):
        pass


class Point(Geometry):
    """Geometric class for point objects.

    Parameters
    ----------
    loc : tuple
        The point's location (number :math:`x`-tuple, :math:`x` > 1).

    Examples
    --------
    >>> p = Point((1, 3))
    """

    def __init__(self, loc):
        warnings.warn(dep_msg, FutureWarning, stacklevel=2)
        self.__loc = tuple(map(float, loc))

    @classmethod
    def __from_geo_interface__(cls, geo):
        return cls(geo["coordinates"])

    @property
    def __geo_interface__(self):
        return {"type": "Point", "coordinates": self.__loc}

    def __lt__(self, other) -> bool:
        """Tests if the point is less than another object.

        Parameters
        ----------
        other : libpysal.cg.Point
            An object to test equality against.

        Examples
        --------
        >>> Point((0, 1)) < Point((0, 1))
        False

        >>> Point((0, 1)) < Point((1, 1))
        True
        """

        return (self.__loc) < (other.__loc)

    def __le__(self, other) -> bool:
        """Tests if the point is less than or equal to another object.

        Parameters
        ----------
        other : libpysal.cg.Point
            An object to test equality against.

        Examples
        --------
        >>> Point((0, 1)) <= Point((0, 1))
        True

        >>> Point((0, 1)) <= Point((1, 1))
        True
        """

        return (self.__loc) <= (other.__loc)

    def __eq__(self, other) -> bool:
        """Tests if the point is equal to another object.

        Parameters
        ----------
        other : libpysal.cg.Point
            An object to test equality against.

        Examples
        --------
        >>> Point((0, 1)) == Point((0, 1))
        True

        >>> Point((0, 1)) == Point((1, 1))
        False
        """

        try:
            return (self.__loc) == (other.__loc)
        except AttributeError:
            return False

    def __ne__(self, other) -> bool:
        """Tests if the point is not equal to another object.

        Parameters
        ----------
        other : libpysal.cg.Point
            An object to test equality against.

        Examples
        --------
        >>> Point((0, 1)) != Point((0, 1))
        False

        >>> Point((0, 1)) != Point((1, 1))
        True
        """

        try:
            return (self.__loc) != (other.__loc)
        except AttributeError:
            return True

    def __gt__(self, other) -> bool:
        """Tests if the point is greater than another object.

        Parameters
        ----------
        other : libpysal.cg.Point
            An object to test equality against.

        Examples
        --------
        >>> Point((0, 1)) > Point((0, 1))
        False

        >>> Point((0, 1)) > Point((1, 1))
        False
        """

        return (self.__loc) > (other.__loc)

    def __ge__(self, other) -> bool:
        """Tests if the point is greater than or equal to another object.

        Parameters
        ----------
        other : libpysal.cg.Point
            An object to test equality against.

        Examples
        --------
        >>> Point((0, 1)) >= Point((0, 1))
        True

        >>> Point((0, 1)) >= Point((1, 1))
        False
        """

        return (self.__loc) >= (other.__loc)

    def __hash__(self) -> int:
        """Returns the hash of the point's location.

        Examples
        --------
        >>> hash(Point((0, 1))) == hash(Point((0, 1)))
        True

        >>> hash(Point((0, 1))) == hash(Point((1, 1)))
        False
        """

        return hash(self.__loc)

    def __getitem__(self, *args) -> int | float:
        """Return the coordinate for the given dimension.

        Parameters
        ----------
        *args : tuple
            A singleton tuple of :math:`(i)` with :math:`i`
            as the index of the desired dimension.

        Examples
        --------
        >>> p = Point((5.5, 4.3))
        >>> p[0] == 5.5
        True
        >>> p[1] == 4.3
        True
        """

        return self.__loc.__getitem__(*args)

    def __getslice__(self, *args) -> slice:
        """Return the coordinates for the given dimensions.

        Parameters
        ----------
        *args : tuple
            A tuple of :math:`(i,j)` with :math:`i` as the index to the start
            slice and :math:`j` as the index to end the slice (excluded).

        Examples
        --------
        >>> p = Point((3, 6, 2))
        >>> p[:2] == (3, 6)
        True

        >>> p[1:2] == (6,)
        True
        """

        return self.__loc.__getslice__(*args)

    def __len__(self) -> int:
        """Returns the dimensions of the point.

        Examples
        --------
        >>> len(Point((1, 2)))
        2
        """

        return len(self.__loc)

    def __repr__(self) -> str:
        """Returns the string representation of the ``Point``.

        Examples
        --------
        >>> Point((0, 1))
        (0.0, 1.0)
        """

        return str(self)

    def __str__(self) -> str:
        """Returns a string representation of a ``Point`` object.

        Examples
        --------
        >>> p = Point((1, 3))
        >>> str(p)
        '(1.0, 3.0)'
        """

        return str(self.__loc)
        # return "POINT ({} {})".format(*self.__loc)


class LineSegment(Geometry):
    """Geometric representation of line segment objects.

    Parameters
    ----------
    start_pt : libpysal.cg.Point
        The point where the segment begins.
    end_pt : libpysal.cg.Point
        The point where the segment ends.

    Attributes
    ----------
    p1 : libpysal.cg.Point
        The starting point of the line segment.
    p2 : Point
        The ending point of the line segment.
    bounding_box : libpysal.cg.Rectangle
        The bounding box of the segment.
    len : float
        The length of the segment.
    line : libpysal.cg.Line
        The line on which the segment lies.

    Examples
    --------
    >>> ls = LineSegment(Point((1, 2)), Point((5, 6)))
    """

    def __init__(self, start_pt, end_pt):
        warnings.warn(dep_msg, FutureWarning, stacklevel=2)
        self._p1 = start_pt
        self._p2 = end_pt
        self._reset_props()

    def __str__(self):
        return "LineSegment(" + str(self._p1) + ", " + str(self._p2) + ")"
        # return "LINESTRING ({} {}, {} {})".format(
        #    self._p1[0], self._p1[1], self._p2[0], self._p2[1]
        # )

    def __eq__(self, other) -> bool:
        """Returns ``True`` if ``self`` and ``other`` are the same line segment.

        Examples
        --------
        >>> l1 = LineSegment(Point((1, 2)), Point((5, 6)))
        >>> l2 = LineSegment(Point((5, 6)), Point((1, 2)))
        >>> l1 == l2
        True

        >>> l2 == l1
        True
        """

        eq = False

        if not isinstance(other, self.__class__):
            pass
        else:
            if (other.p1 == self._p1 and other.p2 == self._p2) or (
                other.p2 == self._p1 and other.p1 == self._p2
            ):
                eq = True

        return eq

    def intersect(self, other) -> bool:
        """Test whether segment intersects with other segment (``True``) or
        not (``False``). Handles endpoints of segments being on other segment.

        Parameters
        ----------
        other : libpysal.cg.LineSegment
            Another line segment to check against.

        Examples
        --------
        >>> ls = LineSegment(Point((5, 0)), Point((10, 0)))
        >>> ls1 = LineSegment(Point((5, 0)), Point((10, 1)))
        >>> ls.intersect(ls1)
        True

        >>> ls2 = LineSegment(Point((5, 1)), Point((10, 1)))
        >>> ls.intersect(ls2)
        False

        >>> ls2 = LineSegment(Point((7, -1)), Point((7, 2)))
        >>> ls.intersect(ls2)
        True
        """

        ccw1 = self.sw_ccw(other.p2)
        ccw2 = self.sw_ccw(other.p1)
        ccw3 = other.sw_ccw(self.p1)
        ccw4 = other.sw_ccw(self.p2)

        intersects = ccw1 * ccw2 <= 0 and ccw3 * ccw4 <= 0

        return intersects

    def _reset_props(self):
        """**HELPER METHOD. DO NOT CALL.**
        Resets attributes which are functions of other attributes.
        The getters for these attributes (implemented as properties)
        then recompute their values if they have been reset since
        the last call to the getter.

        Examples
        --------
        >>> ls = LineSegment(Point((1, 2)), Point((5, 6)))
        >>> ls._reset_props()
        """

        self._bounding_box = None
        self._len = None
        self._line = False

    def _get_p1(self):
        """**HELPER METHOD. DO NOT CALL.**
        Returns the ``p1`` attribute of the line segment.

        Returns
        -------
        self._p1 : libpysal.cg.Point
            The ``_p1`` attribute.

        Examples
        --------
        >>> ls = LineSegment(Point((1, 2)), Point((5, 6)))
        >>> r = ls._get_p1()
        >>> r == Point((1, 2))
        True
        """

        return self._p1

    def _set_p1(self, p1):
        """**HELPER METHOD. DO NOT CALL.**
        Sets the ``p1`` attribute of the line segment.

        Parameters
        ----------
        p1 : libpysal.cg.Point
            A point.

        Returns
        -------
        self._p1 : libpysal.cg.Point
            The reset ``p1`` attribute.

        Examples
        --------
        >>> ls = LineSegment(Point((1, 2)), Point((5, 6)))
        >>> r = ls._set_p1(Point((3, -1)))
        >>> r == Point((3.0, -1.0))
        True
        """

        self._p1 = p1
        self._reset_props()

        return self._p1

    p1 = property(_get_p1, _set_p1)

    def _get_p2(self):
        """**HELPER METHOD. DO NOT CALL.**
        Returns the ``p2`` attribute of the line segment.

        Returns
        -------
        self._p2 : libpysal.cg.Point
            The ``_p2`` attribute.

        Examples
        --------
        >>> ls = LineSegment(Point((1, 2)), Point((5, 6)))
        >>> r = ls._get_p2()
        >>> r == Point((5, 6))
        True
        """

        return self._p2

    def _set_p2(self, p2):
        """**HELPER METHOD. DO NOT CALL.**
        Sets the ``p2`` attribute of the line segment.

        Parameters
        ----------
        p2 : libpysal.cg.Point
            A point.

        Returns
        -------
        self._p2 : libpysal.cg.Point
            The reset ``p2`` attribute.

        Examples
        --------
        >>> ls = LineSegment(Point((1, 2)), Point((5, 6)))
        >>> r = ls._set_p2(Point((3, -1)))
        >>> r == Point((3.0, -1.0))
        True
        """

        self._p2 = p2
        self._reset_props()

        return self._p2

    p2 = property(_get_p2, _set_p2)

    def is_ccw(self, pt) -> bool:
        """Returns whether a point is counterclockwise of the
        segment (``True``) or not (``False``). Exclusive.

        Parameters
        ----------
        pt : libpysal.cg.Point
            A point lying ccw or cw of a segment.

        Examples
        --------
        >>> ls = LineSegment(Point((0, 0)), Point((5, 0)))
        >>> ls.is_ccw(Point((2, 2)))
        True

        >>> ls.is_ccw(Point((2, -2)))
        False
        """

        v1 = (self._p2[0] - self._p1[0], self._p2[1] - self._p1[1])
        v2 = (pt[0] - self._p1[0], pt[1] - self._p1[1])

        return v1[0] * v2[1] - v1[1] * v2[0] > 0

    def is_cw(self, pt) -> bool:
        """Returns whether a point is clockwise of the
        segment (``True``) or not (``False``). Exclusive.

        Parameters
        ----------
        pt : libpysal.cg.Point
            A point lying ccw or cw of a segment.

        Examples
        --------
        >>> ls = LineSegment(Point((0, 0)), Point((5, 0)))
        >>> ls.is_cw(Point((2, 2)))
        False

        >>> ls.is_cw(Point((2, -2)))
        True
        """

        v1 = (self._p2[0] - self._p1[0], self._p2[1] - self._p1[1])
        v2 = (pt[0] - self._p1[0], pt[1] - self._p1[1])

        return v1[0] * v2[1] - v1[1] * v2[0] < 0

    def sw_ccw(self, pt):
        """Sedgewick test for ``pt`` being ccw of segment.

        Returns
        -------
        is_ccw : bool
            ``1`` if turn from ``self.p1`` to ``self.p2`` to ``pt`` is ccw.
            ``-1`` if turn from ``self.p1`` to ``self.p2`` to ``pt`` is cw.
            ``-1`` if the points are collinear and ``self.p1`` is in the middle.
            ``1`` if the points are collinear and ``self.p2`` is in the middle.
            ``0`` if the points are collinear and ``pt`` is in the middle.
        """

        p0 = self.p1
        p1 = self.p2
        p2 = pt

        dx1 = p1[0] - p0[0]
        dy1 = p1[1] - p0[1]
        dx2 = p2[0] - p0[0]
        dy2 = p2[1] - p0[1]

        if dy1 * dx2 < dy2 * dx1:
            is_ccw = 1
        elif (dy1 * dx2 > dy2 * dx1) or (dx1 * dx2 < 0 or dy1 * dy2 < 0):
            is_ccw = -1
        elif dx1 * dx1 + dy1 * dy1 >= dx2 * dx2 + dy2 * dy2:
            is_ccw = 0
        else:
            is_ccw = 1

        return is_ccw

    def get_swap(self):
        """Returns a ``LineSegment`` object which has its endpoints swapped.

        Returns
        -------
        line_seg : libpysal.cg.LineSegment
            The ``LineSegment`` object which has its endpoints swapped.

        Examples
        --------
        >>> ls = LineSegment(Point((1, 2)), Point((5, 6)))
        >>> swap = ls.get_swap()
        >>> swap.p1[0]
        5.0

        >>> swap.p1[1]
        6.0

        >>> swap.p2[0]
        1.0

        >>> swap.p2[1]
        2.0
        """

        line_seg = LineSegment(self._p2, self._p1)

        return line_seg

    @property
    def bounding_box(self):
        """Returns the minimum bounding box of a ``LineSegment`` object.

        Returns
        -------
        self._bounding_box : libpysal.cg.Rectangle
            The bounding box of the line segment.

        Examples
        --------
        >>> ls = LineSegment(Point((1, 2)), Point((5, 6)))
        >>> ls.bounding_box.left
        1.0

        >>> ls.bounding_box.lower
        2.0

        >>> ls.bounding_box.right
        5.0

        >>> ls.bounding_box.upper
        6.0
        """

        # If LineSegment attributes p1, p2 changed, recompute
        if self._bounding_box is None:
            self._bounding_box = Rectangle(
                min([self._p1[0], self._p2[0]]),
                min([self._p1[1], self._p2[1]]),
                max([self._p1[0], self._p2[0]]),
                max([self._p1[1], self._p2[1]]),
            )
        return Rectangle(
            self._bounding_box.left,
            self._bounding_box.lower,
            self._bounding_box.right,
            self._bounding_box.upper,
        )

    @property
    def len(self) -> float:
        """Returns the length of a ``LineSegment`` object.

        Examples
        --------
        >>> ls = LineSegment(Point((2, 2)), Point((5, 2)))
        >>> ls.len
        3.0
        """

        # If LineSegment attributes p1, p2 changed, recompute
        if self._len is None:
            self._len = math.hypot(self._p1[0] - self._p2[0], self._p1[1] - self._p2[1])

        return self._len

    @property
    def line(self):
        """Returns a ``Line`` object of the line on which the segment lies.

        Returns
        -------
        self._line : libpysal.cg.Line
            The ``Line`` object of the line on which the segment lies.

        Examples
        --------
        >>> ls = LineSegment(Point((2, 2)), Point((3, 3)))
        >>> l = ls.line
        >>> l.m
        1.0

        >>> l.b
        0.0
        """

        if self._line is False:
            dx = self._p1[0] - self._p2[0]
            dy = self._p1[1] - self._p2[1]

            if dx == 0 and dy == 0:
                self._line = None
            elif dx == 0:
                self._line = VerticalLine(self._p1[0])
            else:
                m = dy / float(dx)
                # y - mx
                b = self._p1[1] - m * self._p1[0]
                self._line = Line(m, b)

        return self._line


class VerticalLine(Geometry):
    """Geometric representation of verticle line objects.

    Parameters
    ----------
    x : {int, float}
        The :math:`x`-intercept of the line. ``x`` is also an attribute.

    Examples
    --------
    >>> ls = VerticalLine(0)
    >>> ls.m
    inf

    >>> ls.b
    nan
    """

    def __init__(self, x):
        warnings.warn(dep_msg, FutureWarning, stacklevel=2)
        self._x = float(x)
        self.m = float("inf")
        self.b = float("nan")

    def x(self, y) -> float:  # noqa: ARG002
        """Returns the :math:`x`-value of the line at a particular :math:`y`-value.

        Parameters
        ----------
        y : {int, float}
            The :math:`y`-value at which to compute :math:`x`.

        Examples
        --------
        >>> l = VerticalLine(0)
        >>> l.x(0.25)
        0.0
        """

        return self._x

    def y(self, x) -> float:  # noqa: ARG002
        """Returns the :math:`y`-value of the line at a particular :math:`x`-value.

        Parameters
        ----------
        x : {int, float}
            The :math:`x`-value at which to compute :math:`y`.

        Examples
        --------
        >>> l = VerticalLine(1)
        >>> l.y(1)
        nan
        """

        return float("nan")


class Line(Geometry):
    """Geometric representation of line objects.

    Parameters
    ----------
    m : {int, float}
        The slope of the line. ``m`` is also an attribute.
    b : {int, float}
        The :math:`y`-intercept of the line. ``b`` is also an attribute.

    Raises
    ------
    ArithmeticError
        Raised when infinity is passed in as the slope.

    Examples
    --------
    >>> ls = Line(1, 0)
    >>> ls.m
    1.0

    >>> ls.b
    0.0
    """

    def __init__(self, m, b):
        warnings.warn(dep_msg, FutureWarning, stacklevel=2)
        if m == float("inf"):
            raise ArithmeticError("Slope cannot be infinite.")

        self.m = float(m)
        self.b = float(b)

    def x(self, y: int | float) -> float:
        """Returns the :math:`x`-value of the line at a particular :math:`y`-value.

        Parameters
        ----------
        y : {int, float}
            The :math:`y`-value at which to compute :math:`x`.

        Raises
        ------
        ArithmeticError
            Raised when ``0.`` is passed in as the slope.

        Examples
        --------
        >>> l = Line(0.5, 0)
        >>> l.x(0.25)
        0.5
        """

        if self.m == 0:
            raise ArithmeticError("Cannot solve for 'x' when slope is zero.")

        return (y - self.b) / self.m

    def y(self, x: int | float) -> float:
        """Returns the :math:`y`-value of the line at a particular :math:`x`-value.

        Parameters
        ----------
        x : {int, float}
            The :math:`x`-value at which to compute :math:`y`.

        Examples
        --------
        >>> l = Line(1, 0)
        >>> l.y(1)
        1.0
        """

        if self.m == 0:
            return self.b

        return self.m * x + self.b


class Ray:
    """Geometric representation of ray objects.

    Parameters
    ----------
    origin : libpysal.cg.Point
        The point where the ray originates.
    second_p :
        The second point specifying the ray (not ``origin``.)

    Attributes
    ----------
    o : libpysal.cg.Point
        The origin (point where ray originates). See ``origin``.
    p : libpysal.cg.Point
        The second point on the ray (not the point where the
        ray originates). See ``second_p``.

    Examples
    --------
    >>> l = Ray(Point((0, 0)), Point((1, 0)))
    >>> str(l.o)
    '(0.0, 0.0)'

    >>> str(l.p)
    '(1.0, 0.0)'
    """

    def __init__(self, origin, second_p):
        warnings.warn(dep_msg, FutureWarning, stacklevel=2)
        self.o = origin
        self.p = second_p


class Chain(Geometry):
    """Geometric representation of a chain, also known as a polyline.

    Parameters
    ----------
    vertices : list
        A point list or list of point lists.

    Attributes
    ----------
    vertices : list
        The list of points of the vertices of the chain in order.
    len : float
        The geometric length of the chain.

    Examples
    --------
    >>> c = Chain([Point((0, 0)), Point((1, 0)), Point((1, 1)), Point((2, 1))])
    """

    def __init__(self, vertices: list):
        warnings.warn(dep_msg, FutureWarning, stacklevel=2)
        if isinstance(vertices[0], list):
            self._vertices = list(vertices)
        else:
            self._vertices = [vertices]
        self._reset_props()

    @classmethod
    def __from_geo_interface__(cls, geo: dict):
        if geo["type"].lower() == "linestring":
            verts = [Point(pt) for pt in geo["coordinates"]]
        elif geo["type"].lower() == "multilinestring":
            verts = [list(map(Point, part)) for part in geo["coordinates"]]
        else:
            raise TypeError(f"{geo!r} is not a Chain.")
        return cls(verts)

    @property
    def __geo_interface__(self) -> dict:
        if len(self.parts) == 1:
            return {"type": "LineString", "coordinates": self.vertices}
        else:
            return {"type": "MultiLineString", "coordinates": self.parts}

    def _reset_props(self):
        """**HELPER METHOD. DO NOT CALL.** Resets attributes which are
        functions of other attributes. The ``getter``s for these attributes
        (implemented as ``properties``) then recompute their values if they
        have been reset since the last call to the ``getter``.
        """

        self._len = None
        self._arclen = None
        self._bounding_box = None

    @property
    def vertices(self) -> list:
        """Returns the vertices of the chain in clockwise order.

        Examples
        --------
        >>> c = Chain([Point((0, 0)), Point((1, 0)), Point((1, 1)), Point((2, 1))])
        >>> verts = c.vertices
        >>> len(verts)
        4
        """

        return sum(list(self._vertices), [])

    @property
    def parts(self) -> list:
        """Returns the parts (lists of ``libpysal.cg.Point`` objects) of the chain.

        Examples
        --------
        >>> c = Chain(
        ...     [
        ...         [Point((0, 0)), Point((1, 0)), Point((1, 1)), Point((0, 1))],
        ...         [Point((2, 1)), Point((2, 2)), Point((1, 2)), Point((1, 1))]
        ...     ]
        ... )
        >>> len(c.parts)
        2
        """

        return [list(part) for part in self._vertices]

    @property
    def bounding_box(self):
        """Returns the bounding box of the chain.

        Returns
        -------
        self._bounding_box : libpysal.cg.Rectangle
            The bounding box of the chain.

        Examples
        --------
        >>> c = Chain([Point((0, 0)), Point((2, 0)), Point((2, 1)), Point((0, 1))])
        >>> c.bounding_box.left
        0.0

        >>> c.bounding_box.lower
        0.0

        >>> c.bounding_box.right
        2.0

        >>> c.bounding_box.upper
        1.0
        """

        if self._bounding_box is None:
            vertices = self.vertices
            self._bounding_box = Rectangle(
                min([v[0] for v in vertices]),
                min([v[1] for v in vertices]),
                max([v[0] for v in vertices]),
                max([v[1] for v in vertices]),
            )

        return self._bounding_box

    @property
    def len(self) -> int:
        """Returns the geometric length of the chain.

        Examples
        --------
        >>> c = Chain([Point((0, 0)), Point((1, 0)), Point((1, 1)), Point((2, 1))])
        >>> c.len
        3.0

        >>> c = Chain(
        ...     [
        ...         [Point((0, 0)), Point((1, 0)), Point((1, 1))],
        ...         [Point((10, 10)), Point((11, 10)), Point((11, 11))]
        ...     ]
        ... )
        >>> c.len
        4.0
        """

        def dist(v1: tuple, v2: tuple) -> int | float:
            return math.hypot(v1[0] - v2[0], v1[1] - v2[1])

        def part_perimeter(p: list) -> int | float:
            return sum([dist(p[i], p[i + 1]) for i in range(len(p) - 1)])

        if self._len is None:
            self._len = sum([part_perimeter(part) for part in self._vertices])

        return self._len

    @property
    def arclen(self) -> int | float:
        """Returns the geometric length of the chain
        computed using 'arcdistance' (meters).
        """

        def part_perimeter(p: list) -> int | float:
            return sum([arcdist(p[i], p[i + 1]) * 1000.0 for i in range(len(p) - 1)])

        if self._arclen is None:
            self._arclen = sum([part_perimeter(part) for part in self._vertices])

        return self._arclen

    @property
    def segments(self) -> list:
        """Returns the segments that compose the chain."""

        return [
            [LineSegment(a, b) for (a, b) in zip(part[:-1], part[1:], strict=True)]
            for part in self._vertices
        ]


class Ring(Geometry):
    """Geometric representation of a linear ring. Linear rings must be
    closed, the first and last point must be the same. Open rings will
    be closed. This class exists primarily as a geometric primitive to
    form complex polygons with multiple rings and holes. The ordering
    of the vertices is ignored and will not be altered.

    Parameters
    ----------
    vertices : list
        A list of vertices.

    Attributes
    ----------
    vertices : list
        A list of points with the vertices of the ring.
    len : int
        The number of vertices.
    perimeter : float
        The geometric length of the perimeter of the ring.
    bounding_box : libpysal.cg.Rectangle
        The bounding box of the ring.
    area : float
        The area enclosed by the ring.
    centroid : {tuple, libpysal.cg.Point}
        The centroid of the ring defined by the 'center of gravity'
        or 'center or mass'.
    _quad_tree_structure : libpysal.cg.QuadTreeStructureSingleRing
        The quad tree structure for the ring. This structure helps
        test if a point is inside the ring.
    """

    def __init__(self, vertices):
        warnings.warn(dep_msg, FutureWarning, stacklevel=2)
        if vertices[0] != vertices[-1]:
            vertices = vertices[:] + vertices[0:1]
            # msg = "Supplied vertices do not form a closed ring, "
            # msg += "the first and last vertices are not the same."
            # raise ValueError(msg)

        self.vertices = tuple(vertices)
        self._perimeter = None
        self._bounding_box = None
        self._area = None
        self._centroid = None
        self._quad_tree_structure = None

    def __len__(self) -> int:
        return len(self.vertices)

    @property
    def len(self) -> int:
        return len(self)

    @staticmethod
    def dist(v1, v2) -> int | float:
        return math.hypot(v1[0] - v2[0], v1[1] - v2[1])

    @property
    def perimeter(self) -> int | float:
        if self._perimeter is None:
            dist = self.dist
            v = self.vertices
            self._perimeter = sum(
                [dist(v[i], v[i + 1]) for i in range(-1, len(self) - 1)]
            )
        return self._perimeter

    @property
    def bounding_box(self):
        """Returns the bounding box of the ring.

        Returns
        -------
        self._bounding_box : libpysal.cg.Rectangle
            The bounding box of the ring.

        Examples
        --------
        >>> r = Ring(
        ...     [
        ...         Point((0, 0)),
        ...         Point((2, 0)),
        ...         Point((2, 1)),
        ...         Point((0, 1)),
        ...         Point((0, 0))
        ...     ]
        ... )

        >>> r.bounding_box.left
        0.0

        >>> r.bounding_box.lower
        0.0

        >>> r.bounding_box.right
        2.0

        >>> r.bounding_box.upper
        1.0
        """

        if self._bounding_box is None:
            vertices = self.vertices
            x = [v[0] for v in vertices]
            y = [v[1] for v in vertices]
            self._bounding_box = Rectangle(min(x), min(y), max(x), max(y))

        return self._bounding_box

    @property
    def area(self) -> int | float:
        """Returns the area of the ring.

        Examples
        --------
        >>> r = Ring(
        ...     [
        ...         Point((0, 0)),
        ...         Point((2, 0)),
        ...         Point((2, 1)),
        ...         Point((0, 1)),
        ...         Point((0, 0))
        ...     ]
        ... )
        >>> r.area
        2.0
        """

        return abs(self.signed_area)

    @property
    def signed_area(self) -> int | float:
        if self._area is None:
            vertices = self.vertices
            x = [v[0] for v in vertices]
            y = [v[1] for v in vertices]
            n = len(self)

            a = 0.0
            for i in range(n - 1):
                a += (x[i] + x[i + 1]) * (y[i] - y[i + 1])
            a = a * 0.5
            self._area = -a

        return self._area

    @property
    def centroid(self):
        """Returns the centroid of the ring.

        Returns
        -------
        self._centroid : libpysal.cg.Point
            The ring's centroid.

        Notes
        -----
        The centroid returned by this method is the geometric centroid.
        Also known as the 'center of gravity' or 'center of mass'.

        Examples
        --------
        >>> r = Ring(
        ...     [
        ...         Point((0, 0)),
        ...         Point((2, 0)),
        ...         Point((2, 1)),
        ...         Point((0, 1)),
        ...         Point((0, 0))
        ...     ]
        ... )
        >>> str(r.centroid)
        '(1.0, 0.5)'
        """

        if self._centroid is None:
            vertices = self.vertices
            x = [v[0] for v in vertices]
            y = [v[1] for v in vertices]
            a = self.signed_area
            n = len(self)
            cx = 0
            cy = 0
            for i in range(n - 1):
                f = x[i] * y[i + 1] - x[i + 1] * y[i]
                cx += (x[i] + x[i + 1]) * f
                cy += (y[i] + y[i + 1]) * f
            cx = 1.0 / (6 * a) * cx
            cy = 1.0 / (6 * a) * cy
            self._centroid = Point((cx, cy))

        return self._centroid

    def build_quad_tree_structure(self):
        """Build the quad tree structure for this polygon. Once
        the structure is built, speed for testing if a point is
        inside the ring will be increased significantly.
        """

        self._quad_tree_structure = QuadTreeStructureSingleRing(self)

    def contains_point(self, point):
        """Point containment using winding number. The implementation is based on
        `this <http://www.engr.colostate.edu/~dga/dga/papers/point_in_polygon.pdf>`_.

        Parameters
        ----------
        point : libpysal.cg.Point
            The point to test for containment.

        Returns
        -------
        point_contained : bool
            ``True`` if ``point`` is contained within the polygon, otherwise ``False``.
        """

        point_contained = False

        if self._quad_tree_structure is None:
            x, y = point

            # bbox checks
            bbleft = x < self.bounding_box.left
            bbright = x > self.bounding_box.right
            bblower = y < self.bounding_box.lower
            bbupper = y > self.bounding_box.upper

            if bbleft or bbright or bblower or bbupper:
                pass
            else:
                rn = len(self.vertices)
                xs = [self.vertices[i][0] - point[0] for i in range(rn)]
                ys = [self.vertices[i][1] - point[1] for i in range(rn)]
                w = 0

                for i in range(len(self.vertices) - 1):
                    yi = ys[i]
                    yj = ys[i + 1]
                    xi = xs[i]
                    xj = xs[i + 1]
                    if yi * yj < 0:
                        r = xi + yi * (xj - xi) / (yi - yj)
                        if r > 0:
                            if yi < 0:
                                w += 1
                            else:
                                w -= 1
                    elif yi == 0 and xi > 0:
                        if yj > 0:
                            w += 0.5
                        else:
                            w -= 0.5
                    elif yj == 0 and xj > 0:
                        if yi < 0:
                            w += 0.5
                        else:
                            w -= 0.5
                if w == 0:
                    pass
                else:
                    point_contained = True
        else:
            point_contained = self._quad_tree_structure.contains_point(point)

        return point_contained


class Polygon(Geometry):
    """Geometric representation of polygon objects.
    Returns a polygon created from the objects specified.

    Parameters
    ----------
    vertices : list
        A list of vertices or a list of lists of vertices.
    holes : list
        A list of sub-polygons to be considered as holes.
        Default is ``None``.

    Attributes
    ----------
    vertices : list
        A list of points with the vertices of the polygon in clockwise order.
    len : int
        The number of vertices including holes.
    perimeter : float
        The geometric length of the perimeter of the polygon.
    bounding_box : libpysal.cg.Rectangle
        The bounding box of the polygon.
    bbox : list
        A list representation of the bounding box in the
        form ``[left, lower, right, upper]``.
    area : float
        The area enclosed by the polygon.
    centroid : tuple
        The 'center of gravity', i.e. the mean point of the polygon.

    Examples
    --------
    >>> p1 = Polygon([Point((0, 0)), Point((1, 0)), Point((1, 1)), Point((0, 1))])
    """

    def __init__(self, vertices, holes=None):
        warnings.warn(dep_msg, FutureWarning, stacklevel=2)
        self._part_rings = []
        self._hole_rings = []

        def clockwise(part: list) -> list:
            if standalone.is_clockwise(part):
                return part[:]
            else:
                return part[::-1]

        vl = list(vertices)
        if isinstance(vl[0], list):
            self._part_rings = list(map(Ring, vertices))
            self._vertices = [clockwise(part) for part in vertices]
        else:
            self._part_rings = [Ring(vertices)]
            self._vertices = [clockwise(vertices)]
        if holes is not None and holes != []:
            if isinstance(holes[0], list):
                self._hole_rings = list(map(Ring, holes))
                self._holes = [clockwise(hole) for hole in holes]
            else:
                self._hole_rings = [Ring(holes)]
                self._holes = [clockwise(holes)]
        else:
            self._holes = [[]]
        self._reset_props()

    @classmethod
    def __from_geo_interface__(cls, geo: dict):
        """While PySAL does not differentiate polygons and multipolygons
        GEOS, Shapely, and geoJSON do. In GEOS, etc, polygons may only
        have a single exterior ring, all other parts are holes.
        MultiPolygons are simply a list of polygons.
        """

        geo_type = geo["type"].lower()
        if geo_type == "multipolygon":
            parts = []
            holes = []
            for polygon in geo["coordinates"]:
                verts = [[Point(pt) for pt in part] for part in polygon]
                parts += verts[0:1]
                holes += verts[1:]
            if not holes:
                holes = None
            return cls(parts, holes)
        else:
            verts = [[Point(pt) for pt in part] for part in geo["coordinates"]]
            return cls(verts[0:1], verts[1:])

    @property
    def __geo_interface__(self) -> dict:
        """Return ``__geo_interface__`` information lookup."""

        if len(self.parts) > 1:
            geo = {
                "type": "MultiPolygon",
                "coordinates": [[part] for part in self.parts],
            }
            if self._holes[0]:
                geo["coordinates"][0] += self._holes
            return geo
        if self._holes[0]:
            return {"type": "Polygon", "coordinates": self._vertices + self._holes}
        else:
            return {"type": "Polygon", "coordinates": self._vertices}

    def _reset_props(self):
        """Resets the geometric properties of the polygon."""
        self._perimeter = None
        self._bounding_box = None
        self._bbox = None
        self._area = None
        self._centroid = None
        self._len = None

    def __len__(self) -> int:
        return self.len

    @property
    def len(self) -> int:
        """Returns the number of vertices in the polygon.

        Examples
        --------
        >>> p1 = Polygon([Point((0, 0)), Point((0, 1)), Point((1, 1)), Point((1, 0))])
        >>> p1.len
        4

        >>> len(p1)
        4
        """

        if self._len is None:
            self._len = len(self.vertices)
        return self._len

    @property
    def vertices(self) -> list:
        """Returns the vertices of the polygon in clockwise order.

        Examples
        --------
        >>> p1 = Polygon([Point((0, 0)), Point((0, 1)), Point((1, 1)), Point((1, 0))])
        >>> len(p1.vertices)
        4
        """

        return sum(list(self._vertices), []) + sum(list(self._holes), [])

    @property
    def holes(self) -> list:
        """Returns the holes of the polygon in clockwise order.

        Examples
        --------
        >>> p = Polygon(
        ...     [Point((0, 0)), Point((10, 0)), Point((10, 10)), Point((0, 10))],
        ...     [Point((1, 2)), Point((2, 2)), Point((2, 1)), Point((1, 1))]
        ... )
        >>> len(p.holes)
        1
        """

        return [list(part) for part in self._holes]

    @property
    def parts(self) -> list:
        """Returns the parts of the polygon in clockwise order.

        Examples
        --------
        >>> p = Polygon(
        ...     [
        ...         [Point((0, 0)), Point((1, 0)), Point((1, 1)), Point((0, 1))],
        ...         [Point((2, 1)), Point((2, 2)), Point((1, 2)), Point((1, 1))]
        ...     ]
        ... )
        >>> len(p.parts)
        2
        """

        return [list(part) for part in self._vertices]

    @property
    def perimeter(self) -> int | float:
        """Returns the perimeter of the polygon.

        Examples
        --------
        >>> p = Polygon([Point((0, 0)), Point((1, 0)), Point((1, 1)), Point((0, 1))])
        >>> p.perimeter
        4.0
        """

        def dist(v1: int | float, v2: int | float) -> float:
            return math.hypot(v1[0] - v2[0], v1[1] - v2[1])

        def part_perimeter(part) -> int | float:
            return sum([dist(part[i], part[i + 1]) for i in range(-1, len(part) - 1)])

        sum_perim = lambda part_type: sum(  # noqa: E731
            [part_perimeter(part) for part in part_type]
        )

        if self._perimeter is None:
            self._perimeter = sum_perim(self._vertices) + sum_perim(self._holes)

        return self._perimeter

    @property
    def bbox(self):
        """Returns the bounding box of the polygon as a list.

        Returns
        -------
        self._bbox : list
            The bounding box of the polygon as a list.

        See Also
        --------
        libpysal.cg.bounding_box
        """

        if self._bbox is None:
            self._bbox = [
                self.bounding_box.left,
                self.bounding_box.lower,
                self.bounding_box.right,
                self.bounding_box.upper,
            ]
        return self._bbox

    @property
    def bounding_box(self):
        """Returns the bounding box of the polygon.

        Returns
        -------
        self._bounding_box : libpysal.cg.Rectangle
            The bounding box of the polygon.

        Examples
        --------
        >>> p = Polygon([Point((0, 0)), Point((2, 0)), Point((2, 1)), Point((0, 1))])
        >>> p.bounding_box.left
        0.0

        >>> p.bounding_box.lower
        0.0

        >>> p.bounding_box.right
        2.0

        >>> p.bounding_box.upper
        1.0
        """

        if self._bounding_box is None:
            vertices = self.vertices
            self._bounding_box = Rectangle(
                min([v[0] for v in vertices]),
                min([v[1] for v in vertices]),
                max([v[0] for v in vertices]),
                max([v[1] for v in vertices]),
            )
        return self._bounding_box

    @property
    def area(self) -> float:
        """Returns the area of the polygon.

        Examples
        --------
        >>> p = Polygon([Point((0, 0)), Point((1, 0)), Point((1, 1)), Point((0, 1))])
        >>> p.area
        1.0

        >>> p = Polygon(
        ...     [Point((0, 0)), Point((10, 0)), Point((10, 10)), Point((0, 10))],
        ...     [Point((2, 1)), Point((2, 2)), Point((1, 2)), Point((1, 1))]
        ... )
        >>> p.area
        99.0
        """

        def part_area(pv: list) -> float:
            __area = 0
            for i in range(-1, len(pv) - 1):
                __area += (pv[i][0] + pv[i + 1][0]) * (pv[i][1] - pv[i + 1][1])
            __area = __area * 0.5
            if __area < 0:
                __area = -area  # noqa: F821
            return __area

        sum_area = lambda part_type: sum(  # noqa: E731
            [part_area(part) for part in part_type]
        )
        _area = sum_area(self._vertices) - sum_area(self._holes)

        return _area

    @property
    def centroid(self) -> tuple:
        """Returns the centroid of the polygon.

        Notes
        -----
        The centroid returned by this method is the geometric
        centroid and respects multipart polygons with holes.
        Also known as the 'center of gravity' or 'center of mass'.

        Examples
        --------
        >>> p = Polygon(
        ...     [Point((0, 0)), Point((10, 0)), Point((10, 10)), Point((0, 10))],
        ...     [Point((1, 1)), Point((1, 2)), Point((2, 2)), Point((2, 1))]
        ... )
        >>> p.centroid
        (5.0353535353535355, 5.0353535353535355)
        """

        cp = [ring.centroid for ring in self._part_rings]
        ap = [ring.area for ring in self._part_rings]
        ch = [ring.centroid for ring in self._hole_rings]
        ah = [-ring.area for ring in self._hole_rings]

        a = ap + ah
        cx = sum([pt[0] * area for pt, area in zip(cp + ch, a, strict=True)]) / sum(a)
        cy = sum([pt[1] * area for pt, area in zip(cp + ch, a, strict=True)]) / sum(a)

        return cx, cy

    def build_quad_tree_structure(self):
        """Build the quad tree structure for this polygon. Once
        the structure is built, speed for testing if a point is
        inside the ring will be increased significantly.
        """

        for ring in self._part_rings:
            ring.build_quad_tree_structure()
        for ring in self._hole_rings:
            ring.build_quad_tree_structure()
        self.is_quad_tree_structure_built = True

    def contains_point(self, point):
        """Test if a polygon contains a point.

        Parameters
        ----------
        point : libpysal.cg.Point
            A point to test for containment.

        Returns
        -------
        contains : bool
            ``True`` if the polygon contains ``point`` otherwise ``False``.

        Examples
        --------
        >>> p = Polygon(
        ...     [Point((0,0)), Point((4,0)), Point((4,5)), Point((2,3)), Point((0,5))]
        ... )
        >>> p.contains_point((3,3))
        1

        >>> p.contains_point((0,6))
        0

        >>> p.contains_point((2,2.9))
        1

        >>> p.contains_point((4,5))
        0

        >>> p.contains_point((4,0))
        0

        Handles holes.

        >>> p = Polygon(
        ...     [Point((0, 0)), Point((0, 10)), Point((10, 10)), Point((10, 0))],
        ...     [Point((2, 2)), Point((4, 2)), Point((4, 4)), Point((2, 4))]
        ... )
        >>> p.contains_point((3.0, 3.0))
        False

        >>> p.contains_point((1.0, 1.0))
        True

        Notes
        -----
        Points falling exactly on polygon edges may yield unpredictable results.
        """

        searching = True

        for ring in self._hole_rings:
            if ring.contains_point(point):
                contains = False
                searching = False
                break

        if searching:
            for ring in self._part_rings:
                if ring.contains_point(point):
                    contains = True
                    searching = False
                    break
            if searching:
                contains = False

        return contains


class Rectangle(Geometry):
    """Geometric representation of rectangle objects.

    Attributes
    ----------
    left : float
        Minimum x-value of the rectangle.
    lower : float
        Minimum y-value of the rectangle.
    right : float
        Maximum x-value of the rectangle.
    upper : float
        Maximum y-value of the rectangle.

    Examples
    --------
    >>> r = Rectangle(-4, 3, 10, 17)
    >>> r.left #minx
    -4.0

    >>> r.lower #miny
    3.0

    >>> r.right #maxx
    10.0

    >>> r.upper #maxy
    17.0
    """

    def __init__(self, left, lower, right, upper):
        warnings.warn(dep_msg, FutureWarning, stacklevel=2)
        if right < left or upper < lower:
            raise ArithmeticError("Rectangle must have positive area.")
        self.left = float(left)
        self.lower = float(lower)
        self.right = float(right)
        self.upper = float(upper)

    def __bool__(self):
        """Rectangles will evaluate to False if they have zero area.
        ``___nonzero__`` is used "to implement truth value
        testing and the built-in operation ``bool()``"
        ``-- http://docs.python.org/reference/datamodel.html

        Examples
        --------
        >>> r = Rectangle(0, 0, 0, 0)
        >>> bool(r)
        False

        >>> r = Rectangle(0, 0, 1, 1)
        >>> bool(r)
        True
        """

        return bool(self.area)

    def __eq__(self, other):
        if other:
            return self[:] == other[:]
        return False

    def __add__(self, other):
        x, y, X, Y = self[:]  # noqa: N806
        x1, y2, X1, Y1 = other[:]  # noqa: N806

        return Rectangle(
            min(self.left, other.left),
            min(self.lower, other.lower),
            max(self.right, other.right),
            max(self.upper, other.upper),
        )

    def __getitem__(self, key):
        """

        Examples
        --------
        >>> r = Rectangle(-4, 3, 10, 17)
        >>> r[:]
        [-4.0, 3.0, 10.0, 17.0]
        """

        l_ = [self.left, self.lower, self.right, self.upper]

        return l_.__getitem__(key)

    def set_centroid(self, new_center):
        """Moves the rectangle center to a new specified point.

        Parameters
        ----------
        new_center : libpysal.cg.Point
            The new location of the centroid of the polygon.

        Examples
        --------
        >>> r = Rectangle(0, 0, 4, 4)
        >>> r.set_centroid(Point((4, 4)))
        >>> r.left
        2.0

        >>> r.right
        6.0

        >>> r.lower
        2.0

        >>> r.upper
        6.0
        """

        shift = (
            new_center[0] - (self.left + self.right) / 2,
            new_center[1] - (self.lower + self.upper) / 2,
        )

        self.left = self.left + shift[0]
        self.right = self.right + shift[0]
        self.lower = self.lower + shift[1]
        self.upper = self.upper + shift[1]

    def set_scale(self, scale):
        """Rescales the rectangle around its center.

        Parameters
        ----------
        scale : int, float
            The ratio of the new scale to the old
            scale (e.g. 1.0 is current size).

        Examples
        --------
        >>> r = Rectangle(0, 0, 4, 4)
        >>> r.set_scale(2)
        >>> r.left
        -2.0
        >>> r.right
        6.0
        >>> r.lower
        -2.0
        >>> r.upper
        6.0
        """

        center = ((self.left + self.right) / 2, (self.lower + self.upper) / 2)

        self.left = center[0] + scale * (self.left - center[0])
        self.right = center[0] + scale * (self.right - center[0])
        self.lower = center[1] + scale * (self.lower - center[1])
        self.upper = center[1] + scale * (self.upper - center[1])

    @property
    def area(self) -> int | float:
        """Returns the area of the Rectangle.

        Examples
        --------
        >>> r = Rectangle(0, 0, 4, 4)
        >>> r.area
        16.0
        """

        return (self.right - self.left) * (self.upper - self.lower)

    @property
    def width(self) -> int | float:
        """Returns the width of the Rectangle.

        Examples
        --------
        >>> r = Rectangle(0, 0, 4, 4)
        >>> r.width
        4.0
        """

        return self.right - self.left

    @property
    def height(self) -> int | float:
        """Returns the height of the Rectangle.

        Examples
        --------
        >>> r = Rectangle(0, 0, 4, 4)
        >>> r.height
        4.0
        """

        return self.upper - self.lower


_geoJSON_type_to_Pysal_type = {  # noqa: N816
    "point": Point,
    "linestring": Chain,
    "multilinestring": Chain,
    "polygon": Polygon,
    "multipolygon": Polygon,
}

# moving this to top breaks unit tests !
from . import standalone
from .polygonQuadTreeStructure import QuadTreeStructureSingleRing



import numpy as np
import pandas
from packaging.version import Version

from ._utils import _resolve_islands


class SetOpsMixin:
    """
    This implements common useful set operations on weights and dunder methods.
    """

    # dunders

    def __le__(self, other):  # <=
        return self.issubgraph(other)

    def __ge__(self, other):  # >=
        return other.issubgraph(self)

    def __lt__(self, other):  # <
        return self.issubgraph(other) & (len(self) < len(other))

    def __gt__(self, other):  # >
        return other.issubgraph(self) & (len(self) > len(other))

    def __eq__(self, other):  # ==
        return self.equals(other)

    def __ne__(self, other):  # !=
        return not self.equals(other)

    def __and__(self, other):  # &
        return self.intersection(other)

    def __or__(self, other):  # |
        return self.union(other)

    def __xor__(self, other):  # ^
        return self.symmetric_difference(other)

    def __iand__(self, other):
        raise TypeError("Graphs are immutable.")

    def __ior__(self, other):
        raise TypeError("Graphs are immutable.")

    def __len__(self):
        return self.n_edges

    # methods

    def intersects(self, right):
        """
        Returns True if left and right share at least one link,
        irrespective of weights value.
        """
        intersection = self._adjacency.index.drop(self.isolates).intersection(
            right._adjacency.index.drop(right.isolates)
        )
        return len(intersection) > 0

    def intersection(self, right):
        """
        Returns a binary Graph, that includes only those neighbor pairs that exist
        in both left and right.
        """
        from .base import Graph

        intersection = self._adjacency.index.drop(self.isolates).intersection(
            right._adjacency.index.drop(right.isolates)
        )
        return Graph.from_arrays(
            *_resolve_islands(
                intersection.get_level_values("focal"),
                intersection.get_level_values("neighbor"),
                self.unique_ids,
                np.ones(intersection.shape[0], dtype=np.int8),
            )
        )

    def symmetric_difference(self, right):
        """
        Filter out links that are in both left and right Graph objects.
        """
        from .base import Graph

        if not (self.unique_ids == right.unique_ids).all():
            raise ValueError(
                "Cannot do symmetric difference of Graphs that are based on "
                "different sets of unique IDs."
            )

        sym_diff = self._adjacency.index.drop(self.isolates).symmetric_difference(
            right._adjacency.index.drop(right.isolates)
        )
        return Graph.from_arrays(
            *_resolve_islands(
                sym_diff.get_level_values("focal"),
                sym_diff.get_level_values("neighbor"),
                self.unique_ids,
                np.ones(sym_diff.shape[0], dtype=np.int8),
            )
        )

    def union(self, right):
        """
        Provide the union of two Graph objects,
        collecing all links that are in either graph.
        """
        from .base import Graph

        if not (self.unique_ids == right.unique_ids).all():
            raise ValueError(
                "Cannot do union of Graphs that are "
                "based on different sets of unique IDs."
            )

        union = self._adjacency.index.drop(self.isolates).union(
            right._adjacency.index.drop(right.isolates)
        )
        return Graph.from_arrays(
            *_resolve_islands(
                union.get_level_values("focal"),
                union.get_level_values("neighbor"),
                self.unique_ids,
                np.ones(union.shape[0], dtype=np.int8),
            )
        )

    def difference(self, right):
        """
        Provide the set difference between the graph on
        the left and the graph on the right. This returns
        all links in the left graph that are not in the right graph.
        """
        from .base import Graph

        diff = self._adjacency.index.drop(self.isolates).difference(
            right._adjacency.index.drop(right.isolates)
        )
        return Graph.from_arrays(
            *_resolve_islands(
                diff.get_level_values("focal"),
                diff.get_level_values("neighbor"),
                self.unique_ids,
                np.ones(diff.shape[0], dtype=np.int8),
            )
        )

    def issubgraph(self, right):
        """
        Return True if every link in the left Graph also occurs in the right Graph.
        This requires both Graphs are labeled equally. Isolates are ignored.
        """
        join = (
            self._adjacency.drop(self.isolates)
            .reset_index(level=1)
            .merge(
                right._adjacency.drop(right.isolates).reset_index(level=1),
                on=("focal", "neighbor"),
                how="outer",
                indicator=True,
            )
        )
        return not (join._merge == "left_only").any()

    def equals(self, right):
        """
        Check that two graphs are identical. This reqiures them to have
        1. the same edge labels and node labels
        2. in the same order
        3. with the same weights

        This is implemented by comparing the underlying adjacency series.

        This is equivalent to checking whether the sorted list of edge tuples
        (focal, neighbor, weight) for the two graphs are the same.
        """
        try:
            pandas.testing.assert_series_equal(
                self._adjacency, right._adjacency, check_dtype=False
            )
        except AssertionError:
            return False
        return True

    def isomorphic(self, right):
        """
        Check that two graphs are isomorphic. This requires that a re-labelling
        can be found to convert one graph into the other graph. Requires networkx.
        """
        try:
            import networkx as nx
        except ImportError:
            raise ImportError(
                "NetworkX is required to check for graph isomorphism"
            ) from None

        nxleft = self.to_networkx()
        nxright = right.to_networkx()

        if not nx.faster_could_be_isomorphic(nxleft, nxright):
            return False
        elif not nx.could_be_isomorphic(nxleft, nxright):
            # https://github.com/networkx/networkx/issues/7038
            if Version(nx.__version__) == Version("3.2"):
                return nx.is_isomorphic(nxleft, nxright)
            return False
        else:
            return nx.is_isomorphic(nxleft, nxright)



# ruff: noqa: B006, N802

import os
import sys
from warnings import warn

import numpy as np
from scipy import sparse

from .util import lat2SW
from .weights import WSP, W

if os.path.basename(sys.argv[0]) in ("pytest", "py.test"):

    def jit(*dec_args, **dec_kwargs):  # noqa: ARG001
        """
        decorator mimicking numba.jit
        """

        def intercepted_function(f, *f_args, **f_kwargs):  # noqa: ARG001
            return f

        return intercepted_function

else:
    from ..common import jit

__all__ = ["da2W", "da2WSP", "w2da", "wsp2da", "testDataArray"]


def da2W(
    da,
    criterion="queen",
    z_value=None,
    coords_labels={},
    k=1,
    include_nodata=False,
    n_jobs=1,
    **kwargs,
):
    """
    Create a W object from xarray.DataArray with an additional
    attribute index containing coordinate values of the raster
    in the form of Pandas.Index/MultiIndex.

    Parameters
    ----------
    da : xarray.DataArray
        Input 2D or 3D DataArray with shape=(z, y, x)
    criterion : {"rook", "queen"}
        Type of contiguity. Default is queen.
    z_value : int/string/float
        Select the z_value of 3D DataArray with multiple layers.
    coords_labels : dictionary
        Pass dimension labels for coordinates and layers if they do not
        belong to default dimensions, which are (band/time, y/lat, x/lon)
        e.g. coords_labels = {"y_label": "latitude",
        "x_label": "longitude", "z_label": "year"}
        Default is {} empty dictionary.
    k : int
        Order of contiguity, this will select all neighbors upto kth order.
        Default is 1.
    include_nodata : boolean
        If True, missing values will be assumed as non-missing when
        selecting higher_order neighbors, Default is False
    n_jobs : int
        Number of cores to be used in the sparse weight construction. If -1,
        all available cores are used. Default is 1.
    **kwargs : keyword arguments
        Optional arguments for :class:`libpysal.weights.W`

    Returns
    -------
    w : libpysal.weights.W
       instance of spatial weights class W with an index attribute

    Notes
    -----
    1. Lower order contiguities are also selected.
    2. Returned object contains `index` attribute that includes a
       `Pandas.MultiIndex` object from the DataArray.

    Examples
    --------
    >>> from libpysal.weights.raster import da2W, testDataArray
    >>> da = testDataArray().rename(
            {'band': 'layer', 'x': 'longitude', 'y': 'latitude'})
    >>> da.dims
    ('layer', 'latitude', 'longitude')
    >>> da.shape
    (3, 4, 4)
    >>> da.coords
    Coordinates:
        * layer      (layer) int64 1 2 3
        * latitude   (latitude) float64 90.0 30.0 -30.0 -90.0
        * longitude  (longitude) float64 -180.0 -60.0 60.0 180.0
    >>> da.attrs
    {'nodatavals': (-32768.0,)}
    >>> coords_labels = {
        "z_label": "layer",
        "y_label": "latitude",
        "x_label": "longitude"
    }
    >>> w = da2W(da, z_value=2, coords_labels=coords_labels)
    >>> "%.3f"%w.pct_nonzero
    '30.000'
    >>> w[(2, 90.0, 180.0)] == {(2, 90.0, 60.0): 1, (2, 30.0, 180.0): 1}
    True
    >>> len(w.index)
    10
    >>> w.index[:2]
    MultiIndex([(2, 90.0,  60.0),
                (2, 90.0, 180.0)],
               names=['layer', 'latitude', 'longitude'])

    See Also
    --------
    :class:`libpysal.weights.weights.W`
    """
    warn(
        "You are trying to build a full W object from "
        "xarray.DataArray (raster) object. This computation "
        "can be very slow and not scale well. It is recommended, "
        "if possible, to instead build WSP object, which is more "
        "efficient and faster. You can do this by using da2WSP method.",
        stacklevel=2,
    )
    wsp = da2WSP(da, criterion, z_value, coords_labels, k, include_nodata, n_jobs)
    w = wsp.to_W(**kwargs)

    # temp addition of index attribute
    w.index = wsp.index
    return w


def da2WSP(
    da,
    criterion="queen",
    z_value=None,
    coords_labels={},
    k=1,
    include_nodata=False,
    n_jobs=1,
):
    """
    Create a WSP object from xarray.DataArray with an additional
    attribute index containing coordinate values of the raster
    in the form of Pandas.Index/MultiIndex.

    Parameters
    ----------
    da : xarray.DataArray
        Input 2D or 3D DataArray with shape=(z, y, x)
    criterion : {"rook", "queen"}
        Type of contiguity. Default is queen.
    z_value : int/string/float
        Select the z_value of 3D DataArray with multiple layers.
    coords_labels : dictionary
        Pass dimension labels for coordinates and layers if they do not
        belong to default dimensions, which are (band/time, y/lat, x/lon)
        e.g. coords_labels = {"y_label": "latitude",
        "x_label": "longitude", "z_label": "year"}
        Default is {} empty dictionary.
    k : int
        Order of contiguity, this will select all neighbors upto kth order.
        Default is 1.
    include_nodata : boolean
        If True, missing values will be assumed as non-missing when
        selecting higher_order neighbors, Default is False
    n_jobs : int
        Number of cores to be used in the sparse weight construction. If -1,
        all available cores are used. Default is 1.

    Returns
    -------
    wsp : libpysal.weights.WSP
       instance of spatial weights class WSP with an index attribute

    Notes
    -----
    1. Lower order contiguities are also selected.
    2. Returned object contains `index` attribute that includes a
       `Pandas.MultiIndex` object from the DataArray.

    Examples
    --------
    >>> from libpysal.weights.raster import da2WSP, testDataArray
    >>> da = testDataArray().rename(
            {'band': 'layer', 'x': 'longitude', 'y': 'latitude'})
    >>> da.dims
    ('layer', 'latitude', 'longitude')
    >>> da.shape
    (3, 4, 4)
    >>> da.coords
    Coordinates:
        * layer      (layer) int64 1 2 3
        * latitude   (latitude) float64 90.0 30.0 -30.0 -90.0
        * longitude  (longitude) float64 -180.0 -60.0 60.0 180.0
    >>> da.attrs
    {'nodatavals': (-32768.0,)}
    >>> coords_labels = {
        "z_label": "layer",
        "y_label": "latitude",
        "x_label": "longitude"
    }
    >>> wsp = da2WSP(da, z_value=2, coords_labels=coords_labels)
    >>> wsp.n
    10
    >>> pct_sp = wsp.sparse.nnz *1. / wsp.n**2
    >>> "%.3f"%pct_sp
    '0.300'
    >>> print(wsp.sparse[4].todense())
    [[0 0 1 0 0 1 1 1 0 0]]
    >>> wsp.index[:2]
    MultiIndex([(2, 90.0,  60.0),
                (2, 90.0, 180.0)],
               names=['layer', 'latitude', 'longitude'])

    See Also
    --------
    :class:`libpysal.weights.weights.WSP`
    """
    try:
        import numba  # noqa: F401

        use_numba = True
    except (ModuleNotFoundError, ImportError):
        warn(
            "numba cannot be imported, parallel processing "
            "and include_nodata functionality will be disabled. "
            "falling back to slower method",
            stacklevel=2,
        )
        use_numba = False
        include_nodata = False

    if use_numba:
        sw_tup, ser, n = _da2wsp(
            da,
            criterion=criterion,
            z_value=z_value,
            coords_labels=coords_labels,
            k=k,
            include_nodata=include_nodata,
            n_jobs=n_jobs,
            use_numba=use_numba,
        )

        sw = sparse.csr_matrix(
            sw_tup,
            shape=(n, n),
            dtype=np.int8,
        )
    else:
        sw, ser = _da2wsp(
            da,
            criterion=criterion,
            z_value=z_value,
            coords_labels=coords_labels,
            k=k,
            include_nodata=include_nodata,
            n_jobs=n_jobs,
            use_numba=use_numba,
        )

    # Higher_order functionality, this uses idea from
    # libpysal#313 for adding higher order neighbors.
    # Since diagonal elements are also added in the result,
    # this method set the diagonal elements to zero and
    # then eliminate zeros from the data. This changes the
    # sparsity of the csr_matrix !!
    if k > 1 and not include_nodata:
        sw = sum(sw**x for x in range(1, k + 1))
        sw.setdiag(0)
        sw.eliminate_zeros()
        sw.data[:] = np.ones_like(sw.data, dtype=np.int8)

    index = ser.index
    wsp = WSP(sw, index=index)
    return wsp


def _da2wsp(
    da,
    criterion="queen",
    z_value=None,
    coords_labels={},
    k=1,
    include_nodata=False,
    n_jobs=1,
    use_numba=True,
):
    """Helper for da2WSP that can be reused in Graph"""
    z_id, coords_labels = _da_checker(da, z_value, coords_labels)
    shape = da.shape
    if z_id:
        slice_dict = {}
        slice_dict[coords_labels["z_label"]] = 0
        shape = da[slice_dict].shape
        slice_dict[coords_labels["z_label"]] = slice(z_id - 1, z_id)
        da = da[slice_dict]

    ser = da.to_series()
    dtype = np.int32 if (shape[0] * shape[1]) < 46340**2 else np.int64
    if "nodatavals" in da.attrs and da.attrs["nodatavals"]:
        mask = (ser != da.attrs["nodatavals"][0]).to_numpy()
        ids = np.where(mask)[0]
        id_map = _idmap(ids, mask, dtype)
        ser = ser[ser != da.attrs["nodatavals"][0]]
    else:
        ids = np.arange(len(ser), dtype=dtype)
        id_map = ids.copy()

    n = len(ids)

    if use_numba:
        k_nas = k if include_nodata else 1

        if n_jobs != 1:
            try:
                import joblib  # noqa: F401
            except (ModuleNotFoundError, ImportError):
                warn(
                    f"Parallel processing is requested (n_jobs={n_jobs}),"
                    f" but joblib cannot be imported. n_jobs will be set"
                    f" to 1.",
                    stacklevel=2,
                )
                n_jobs = 1

        if n_jobs == 1:
            sw_tup = _SWbuilder(
                *shape, ids, id_map, criterion, k_nas, dtype
            )  # -> (data, (row, col))
        else:
            if n_jobs == -1:
                n_jobs = os.cpu_count()
            # Parallel implementation
            sw_tup = _parSWbuilder(
                *shape, ids, id_map, criterion, k_nas, dtype, n_jobs
            )  # -> (data, (row, col))
    else:
        # Fallback method to build sparse matrix
        sw = lat2SW(*shape, criterion)
        if "nodatavals" in da.attrs and da.attrs["nodatavals"]:
            sw = sw[mask]
            sw = sw[:, mask]
        return sw, ser
    return sw_tup, ser, n


def w2da(data, w, attrs={}, coords=None):
    """
    Creates xarray.DataArray object from passed data aligned with W object.

    Parameters
    ----------
    data : array/list/pd.Series
        1d array-like data with dimensionality conforming to w
    w : libpysal.weights.W
        Spatial weights object aligned with passed data
    attrs : Dictionary
        Attributes stored in dict related to DataArray, e.g. da.attrs
        Default is {} empty dictionary.
    coords : Dictionary/xarray.core.coordinates.DataArrayCoordinates
        Coordinates corresponding to DataArray, e.g. da.coords

    Returns
    -------
    da : xarray.DataArray
        instance of xarray.DataArray

    Examples
    --------
    >>> from libpysal.raster import da2W, testDataArray, w2da
    >>> da = testDataArray()
    >>> da.shape
    (3, 4, 4)
    >>> w = da2W(da, z_value=2)
    >>> data = np.random.randint(0, 255, len(w.index))
    >>> da1 = w2da(data, w)
    """
    if not isinstance(w, W):
        raise TypeError("w must be an instance of weights.W")
    if hasattr(w, "index"):
        da = _index2da(data, w.index, attrs, coords)
    else:
        raise AttributeError(
            "This method requires `w` object to include `index` "
            "attribute that is built as a `pandas.MultiIndex` object."
        )
    return da


def wsp2da(data, wsp, attrs={}, coords=None):
    """
    Creates xarray.DataArray object from passed data aligned with WSP object.

    Parameters
    ----------
    data : array/list/pd.Series
        1d array-like data with dimensionality conforming to wsp
    wsp : libpysal.weights.WSP
        Sparse weights object aligned with passed data
    attrs : Dictionary
        Attributes stored in dict related to DataArray, e.g. da.attrs
        Default is {} empty dictionary.
    coords : Dictionary/xarray.core.coordinates.DataArrayCoordinates
        coordinates corresponding to DataArray, e.g. da.coords

    Returns
    -------
    da : xarray.DataArray
        instance of xarray.DataArray

    Examples
    --------
    >>> from libpysal.raster import da2WSP, testDataArray, wsp2da
    >>> da = testDataArray()
    >>> da.shape
    (3, 4, 4)
    >>> wsp = da2WSP(da, z_value=2)
    >>> data = np.random.randint(0, 255, len(wsp.index))
    >>> da1 = w2da(data, wsp)
    """
    if not isinstance(wsp, WSP):
        raise TypeError("wsp must be an instance of weights.WSP")
    if hasattr(wsp, "index"):
        da = _index2da(data, wsp.index, attrs, coords)
    else:
        raise AttributeError(
            "This method requires `wsp` object to include `index` "
            "attribute that is built as a `pandas.MultiIndex` object."
        )
    return da


def testDataArray(shape=(3, 4, 4), time=False, rand=False, missing_vals=True):
    """
    Creates 2 or 3 dimensional test xarray.DataArray object

    Parameters
    ----------
    shape : tuple
        Tuple containing shape of the DataArray aligned with
        following dimension = (lat, lon) or (layer, lat, lon)
        Default shape = (3, 4, 4)
    time : boolean
        Type of layer, if True then layer=time else layer=band
        Default is False.
    rand : boolean
        If True, creates a DataArray filled with unique and random data.
        Default is false (generates seeded random data)
    missing_vals : boolean
        Create a DataArray filled with missing values. Default is True.

    Returns
    -------
    da : xarray.DataArray
        instance of xarray.DataArray
    """
    try:
        from xarray import DataArray
    except ImportError:
        raise ModuleNotFoundError(
            "xarray must be installed to use this functionality"
        ) from None
    if not rand:
        np.random.seed(12345)
    coords = {}
    n = len(shape)
    if n != 2:
        layer = "time" if time else "band"
        dims = (layer, "y", "x")
        if time:
            layers = np.arange(
                np.datetime64("2020-07-30"), shape[0], dtype="datetime64[D]"
            )
        else:
            layers = np.arange(1, shape[0] + 1)
        coords[dims[-3]] = layers
    else:
        dims = ("y", "x")
    coords[dims[-2]] = np.linspace(90, -90, shape[-2])
    coords[dims[-1]] = np.linspace(-180, 180, shape[-1])
    data = np.random.randint(0, 255, shape)
    attrs = {}
    if missing_vals:
        attrs["nodatavals"] = (-32768.0,)
        miss_ids = np.where(np.random.randint(2, size=shape) == 1)
        data[miss_ids] = attrs["nodatavals"][0]
    da = DataArray(data, coords, dims, attrs=attrs)
    return da


def _da_checker(da, z_value, coords_labels):
    """
    xarray.dataarray checker for raster interface

    Parameters
    ----------
    da : xarray.DataArray
        Input 2D or 3D DataArray with shape=(z, y, x)
    z_value : int/string/float
        Select the z_value of 3D DataArray with multiple layers.
    coords_labels : dictionary
        Pass dimension labels for coordinates and layers if they do not
        belong to default dimensions, which are (band/time, y/lat, x/lon)
        e.g. coords_labels = {"y_label": "latitude","
        "x_label": "longitude", "z_label": "year"}
        Default is {} empty dictionary.

    Returns
    -------
    z_id : int
        Returns the index of layer
    dims : dictionary
        Mapped dimensions of the DataArray
    """
    try:
        from xarray import DataArray
    except ImportError:
        raise ModuleNotFoundError(
            "xarray must be installed to use this functionality"
        ) from None

    if not isinstance(da, DataArray):
        raise TypeError("da must be an instance of xarray.DataArray")
    if da.ndim not in [2, 3]:
        raise ValueError("da must be 2D or 3D")
    if not (
        np.issubdtype(da.values.dtype, np.integer)
        or np.issubdtype(da.values.dtype, np.floating)
    ):
        raise ValueError("da must be an array of integers or float")

    # default dimensions
    def_labels = {
        "x_label": coords_labels["x_label"]
        if "x_label" in coords_labels
        else ("x" if hasattr(da, "x") else "lon"),
        "y_label": coords_labels["y_label"]
        if "y_label" in coords_labels
        else ("y" if hasattr(da, "y") else "lat"),
    }

    if da.ndim == 3:
        def_labels["z_label"] = (
            coords_labels["z_label"]
            if "z_label" in coords_labels
            else ("band" if hasattr(da, "band") else "time")
        )

        z_id = 1
        if z_value is None:
            if da.sizes[def_labels["z_label"]] != 1:
                warn(
                    "Multiple layers detected. Using first layer as default.",
                    stacklevel=2,
                )
        else:
            z_id += tuple(da[def_labels["z_label"]]).index(z_value)
    else:
        z_id = None
    return z_id, def_labels


def _index2da(data, index, attrs, coords):
    """
    Creates xarray.DataArray object from passed data

    Parameters
    ----------
    data : array/list/pd.Series
        1d array-like data with dimensionality conforming to index
    index : pd.MultiIndex
        indices of the DataArray when converted to pd.Series
    attrs : Dictionary
        Attributes stored in dict related to DataArray, e.g. da.attrs
    coords : Dictionary/xarray.core.coordinates.DataArrayCoordinates
        coordinates corresponding to DataArray, e.g. da[n-1:n].coords

    Returns
    -------
    da : xarray.DataArray
        instance of xarray.DataArray
    """
    try:
        from xarray import DataArray
    except ImportError:
        raise ModuleNotFoundError(
            "xarray must be installed to use this functionality"
        ) from None

    data = np.array(data).flatten()
    idx = index
    dims = idx.names
    indexer = tuple(idx.codes)
    shape = tuple(lev.size for lev in idx.levels)

    if coords is None:
        missing = np.prod(shape) > idx.shape[0]
        if missing:
            if "nodatavals" in attrs:
                fill_value = attrs["nodatavals"][0]
            else:
                min_data = np.min(data)
                fill_value = min_data - 1 if min_data < 0 else -1
                attrs["nodatavals"] = tuple([fill_value])  # noqa: C409
            data_complete = np.full(shape, fill_value, data.dtype)
        else:
            data_complete = np.empty(shape, data.dtype)
        data_complete[indexer] = data
        coords = {}
        for dim, lev in zip(dims, idx.levels, strict=True):
            coords[dim] = lev.to_numpy()
    else:
        fill = attrs["nodatavals"][0] if "nodatavals" in attrs else 0
        data_complete = np.full(shape, fill, data.dtype)
        data_complete[indexer] = data

    da = DataArray(data_complete, coords=coords, dims=dims, attrs=attrs)
    return da


@jit(nopython=True, fastmath=True)
def _idmap(ids, mask, dtype):
    """
    Utility function computes id_map of non-missing raster data

    Parameters
    ----------
    ids : ndarray
        1D array containing ids of non-missing raster data
    mask : ndarray
        1D array mask array
    dtype : type
        Data type of the id_map array

    Returns
    -------
    id_map : ndarray
        1D array containing id_maps of non-missing raster data
    """
    id_map = mask * 1
    id_map[ids] = np.arange(len(ids), dtype=dtype)
    return id_map


@jit(nopython=True, fastmath=True)
def _SWbuilder(
    nrows,
    ncols,
    ids,
    id_map,
    criterion,
    k,
    dtype,
):
    """
    Computes data and orders rows, cols, data for a single chunk

    Parameters
    ----------
    nrows : int
        Number of rows in the raster data
    ncols : int
        Number of columns in the raster data
    ids : ndarray
        1D array containing ids of non-missing raster data
    id_map : ndarray
        1D array containing id_maps of non-missing raster data
    criterion : str
        Type of contiguity.
    k : int
        Order of contiguity, Default is 1
    dtype : type
        Data type of the id_map array

    Returns
    -------
    data : ndarray
        1D ones array containing weight of each neighbor
    rows : ndarray
        1D ones array containing row value of each id
        in the sparse weight object
    cols : ndarray
        1D ones array containing columns value of each id
        in the sparse weight object
    """
    rows, cols = _compute_chunk(nrows, ncols, ids, id_map, criterion, k, dtype)
    data = np.ones_like(rows, dtype=np.int8)
    return (data, (rows, cols))


@jit(nopython=True, fastmath=True, nogil=True)
def _compute_chunk(
    nrows,
    ncols,
    ids,
    id_map,
    criterion,
    k,
    dtype,
):
    """
    Computes rows cols for a single chunk

    Parameters
    ----------
    nrows : int
        Number of rows in the raster data
    ncols : int
        Number of columns in the raster data
    ids : ndarray
        1D array containing ids of non-missing raster data
    id_map : ndarray
        1D array containing id_maps of non-missing raster data
    criterion : str
        Type of contiguity.
    k : int
        Order of contiguity, Default is 1
    dtype : type
        Data type of the rows and cols array

    Returns
    -------
    rows : ndarray
        1D ones array containing row value of each id
        in the sparse weight object
    cols : ndarray
        1D ones array containing columns value of each id
        in the sparse weight object
    ni : int
        Number of rows and cols
    """
    n = len(ids)
    # Setting d which is used for row, col preallocation
    d = 4 if criterion == "rook" else 8
    if k > 1:
        d = int((k / 2) * (2 * 8 + (k - 1) * 8))
    rows = np.empty(d * n, dtype=dtype)
    cols = np.empty_like(rows)
    ni = 0  # -> Pointer to store rows and cols in array
    for order in range(1, k + 1):
        condition = (
            (order - 1)
            if criterion == "queen"
            else ((k - order) if ((k - order) < order) else (order - 1))
        )
        for i in range(n):
            id_i = ids[i]
            og_id = id_map[id_i]

            if ((id_i + order) % ncols) >= order:
                # east neighbor
                id_neighbor = id_map[id_i + order]
                if id_neighbor:
                    rows[ni], cols[ni] = og_id, id_neighbor
                    ni += 1
                    rows[ni], cols[ni] = id_neighbor, og_id
                    ni += 1
                # north-east to south-east neighbors
                for j in range(condition):
                    if (id_i // ncols) < (nrows - j - 1):
                        id_neighbor = id_map[(id_i + order) + (ncols * (j + 1))]
                        if id_neighbor:
                            rows[ni], cols[ni] = og_id, id_neighbor
                            ni += 1
                            rows[ni], cols[ni] = id_neighbor, og_id
                            ni += 1
                    if (id_i // ncols) >= j + 1:
                        id_neighbor = id_map[(id_i + order) - (ncols * (j + 1))]
                        if id_neighbor:
                            rows[ni], cols[ni] = og_id, id_neighbor
                            ni += 1
                            rows[ni], cols[ni] = id_neighbor, og_id
                            ni += 1

            if (id_i // ncols) < (nrows - order):
                # south neighbor
                id_neighbor = id_map[id_i + (ncols * order)]
                if id_neighbor:
                    rows[ni], cols[ni] = og_id, id_neighbor
                    ni += 1
                    rows[ni], cols[ni] = id_neighbor, og_id
                    ni += 1
                # south-west to south-east neighbors
                for j in range(condition):
                    if (id_i % ncols) >= j + 1:
                        id_neighbor = id_map[id_i + (ncols * order) - j - 1]
                        if id_neighbor:
                            rows[ni], cols[ni] = og_id, id_neighbor
                            ni += 1
                            rows[ni], cols[ni] = id_neighbor, og_id
                            ni += 1
                    if ((id_i + j + 1) % ncols) >= j + 1:
                        id_neighbor = id_map[id_i + (ncols * order) + j + 1]
                        if id_neighbor:
                            rows[ni], cols[ni] = og_id, id_neighbor
                            ni += 1
                            rows[ni], cols[ni] = id_neighbor, og_id
                            ni += 1

                if criterion == "queen" or ((k / order) >= 2.0):
                    if (id_i % ncols) >= order:
                        # south-west neighbor
                        id_neighbor = id_map[id_i + (ncols * order) - order]
                        if id_neighbor:
                            rows[ni], cols[ni] = og_id, id_neighbor
                            ni += 1
                            rows[ni], cols[ni] = id_neighbor, og_id
                            ni += 1
                    if ((id_i + order) % ncols) >= order:
                        # south-east neighbor
                        id_neighbor = id_map[id_i + (ncols * order) + order]
                        if id_neighbor:
                            rows[ni], cols[ni] = og_id, id_neighbor
                            ni += 1
                            rows[ni], cols[ni] = id_neighbor, og_id
                            ni += 1
    return rows[:ni], cols[:ni]


@jit(nopython=True, fastmath=True)
def _chunk_generator(
    n_jobs,
    starts,
    ids,
):
    """
    Construct chunks to iterate over within numba in parallel

    Parameters
    ----------
    n_jobs : int
        Number of cores to be used in the sparse weight construction. If -1,
        all available cores are used.
    starts : ndarray
        (n_chunks+1,) array of positional starts for ids chunk
    ids : ndarray
        1D array containing ids of non-missing raster data

    Yields
    ------
    ids_chunk : numpy.ndarray
        (n_chunk,) array containing the chunk of non-missing raster data
    """
    chunk_size = starts[1] - starts[0]
    for i in range(n_jobs):
        start = starts[i]
        ids_chunk = ids[start : (start + chunk_size)]
        yield (ids_chunk,)


def _parSWbuilder(
    nrows,
    ncols,
    ids,
    id_map,
    criterion,
    k,
    dtype,
    n_jobs,
):
    """
    Computes data and orders rows, cols, data in parallel using numba

    Parameters
    ----------
    nrows : int
        Number of rows in the raster data
    ncols : int
        Number of columns in the raster data
    ids : ndarray
        1D array containing ids of non-missing raster data
    id_map : ndarray
        1D array containing id_maps of non-missing raster data
    criterion : str
        Type of contiguity.
    k : int
        Order of contiguity, Default is 1
    dtype : type
        Data type of the rows and cols array
    n_jobs : int
        Number of cores to be used in the sparse weight construction. If -1,
        all available cores are used.

    Returns
    -------
    data : ndarray
        1D ones array containing weight of each neighbor
    rows : ndarray
        1D ones array containing row value of each id
        in the sparse weight object
    cols : ndarray
        1D ones array containing columns value of each id
        in the sparse weight object
    """
    from joblib import Parallel, delayed, parallel_backend

    n = len(ids)
    chunk_size = n // n_jobs + 1
    starts = np.arange(n_jobs + 1) * chunk_size
    chunk = _chunk_generator(n_jobs, starts, ids)
    with parallel_backend("threading"):
        worker_out = Parallel(n_jobs=n_jobs)(
            delayed(_compute_chunk)(nrows, ncols, *ids, id_map, criterion, k, dtype)
            for ids in chunk
        )
    rows, cols = zip(*worker_out, strict=True)
    rows = np.concatenate(rows)
    cols = np.concatenate(cols)
    data = np.ones_like(rows, dtype=np.int8)
    return (data, (rows, cols))



def _build_from_h3(ids, order=1):
    """Generate Graph from H3 hexagons.

    Encode a graph from a set of H3 hexagons. The graph is generated by
    considering the H3 hexagons as nodes and connecting them based on their
    contiguity. The contiguity is defined by the order parameter, which
    specifies the number of steps to consider as neighbors.

    Requires the `h3` library.

    Parameters
    ----------
    ids : array-like
        Array of H3 IDs encoding focal geometries
    order : int, optional
        Order of contiguity, by default 1

    Returns
    -------
    tuple(dict, dict)
    """
    try:
        import h3
    except ImportError as e:
        raise ImportError(
            "This function requires the `h3` library. "
            "You can install it with `conda install h3-py` or "
            "`pip install h3`."
        ) from e

    neighbors = {}
    weights = {}
    for ix in ids:
        rings = h3.hex_range_distances(ix, order)
        for i, ring in enumerate(rings):
            if i == 0:
                neighbors[ix] = []
                weights[ix] = []
            else:
                neighbors[ix].extend(list(ring))
                weights[ix].extend([i] * len(ring))

    return neighbors, weights



from shapely import __version__ as shapely_version
from shapely import geometry as geom
from shapely import ops as shops

from .shapes import asShape

_basegeom = geom.base.BaseGeometry


__all__ = [
    "to_wkb",
    "to_wkt",
    "area",
    "distance",
    "length",
    "boundary",
    "bounds",
    "centroid",
    "representative_point",
    "convex_hull",
    "envelope",
    "buffer",
    "simplify",
    "difference",
    "intersection",
    "symmetric_difference",
    "union",
    "unary_union",
    "cascaded_union",
    "has_z",
    "is_empty",
    "is_ring",
    "is_simple",
    "is_valid",
    "relate",
    "contains",
    "crosses",
    "disjoint",
    "equals",
    "intersects",
    "overlaps",
    "touches",
    "within",
    "equals_exact",
    "almost_equals",
    "project",
    "interpolate",
]


GEO_INTERFACE_ATTR = "__geo_interface__"
SHAPE_TYPE_ERR = "%r does not appear to be a shape."


def to_wkb(shape):
    if not hasattr(shape, GEO_INTERFACE_ATTR):
        raise TypeError(SHAPE_TYPE_ERR % shape)
    o = geom.shape(shape)
    return o.to_wkb()


def to_wkt(shape):
    if not hasattr(shape, GEO_INTERFACE_ATTR):
        raise TypeError(SHAPE_TYPE_ERR % shape)
    o = geom.shape(shape)
    return o.to_wkt()


# Real-valued properties and methods
# ----------------------------------
def area(shape):
    if not hasattr(shape, GEO_INTERFACE_ATTR):
        raise TypeError(SHAPE_TYPE_ERR % shape)
    o = geom.shape(shape)
    return o.area


def distance(shape, other):
    if not hasattr(shape, GEO_INTERFACE_ATTR):
        raise TypeError(SHAPE_TYPE_ERR % shape)
    if not hasattr(other, GEO_INTERFACE_ATTR):
        raise TypeError(SHAPE_TYPE_ERR % shape)
    o = geom.shape(shape)
    o2 = geom.shape(other)
    return o.distance(o2)


def length(shape):
    if not hasattr(shape, GEO_INTERFACE_ATTR):
        raise TypeError(SHAPE_TYPE_ERR % shape)
    o = geom.shape(shape)
    return o.length


# Topological properties
# ----------------------
def boundary(shape):
    if not hasattr(shape, GEO_INTERFACE_ATTR):
        raise TypeError(SHAPE_TYPE_ERR % shape)
    o = geom.shape(shape)
    res = o.boundary
    return asShape(res)


def bounds(shape):
    if not hasattr(shape, GEO_INTERFACE_ATTR):
        raise TypeError(SHAPE_TYPE_ERR % shape)
    o = geom.shape(shape)
    return o.bounds


def centroid(shape):
    if not hasattr(shape, GEO_INTERFACE_ATTR):
        raise TypeError(SHAPE_TYPE_ERR % shape)
    o = geom.shape(shape)
    res = o.centroid
    return asShape(res)


def representative_point(shape):
    if not hasattr(shape, GEO_INTERFACE_ATTR):
        raise TypeError(SHAPE_TYPE_ERR % shape)
    o = geom.shape(shape)
    res = o.representative_point()
    return asShape(res)


def convex_hull(shape):
    if not hasattr(shape, GEO_INTERFACE_ATTR):
        raise TypeError(SHAPE_TYPE_ERR % shape)
    o = geom.shape(shape)
    res = o.convex_hull
    return asShape(res)


def envelope(shape):
    if not hasattr(shape, GEO_INTERFACE_ATTR):
        raise TypeError(SHAPE_TYPE_ERR % shape)
    o = geom.shape(shape)
    res = o.envelope
    return asShape(res)


def buffer(shape, radius, resolution=16):
    if not hasattr(shape, GEO_INTERFACE_ATTR):
        raise TypeError(SHAPE_TYPE_ERR % shape)
    o = geom.shape(shape)
    res = o.buffer(radius, resolution)
    return asShape(res)


def simplify(shape, tolerance, preserve_topology=True):
    if not hasattr(shape, GEO_INTERFACE_ATTR):
        raise TypeError(SHAPE_TYPE_ERR % shape)
    o = geom.shape(shape)
    res = o.simplify(tolerance, preserve_topology)
    return asShape(res)


# Binary operations
# -----------------
def difference(shape, other):
    if not hasattr(shape, GEO_INTERFACE_ATTR):
        raise TypeError(SHAPE_TYPE_ERR % shape)
    if not hasattr(other, GEO_INTERFACE_ATTR):
        raise TypeError(SHAPE_TYPE_ERR % shape)
    o = geom.shape(shape)
    o2 = geom.shape(other)
    res = o.difference(o2)
    return asShape(res)


def intersection(shape, other):
    if not hasattr(shape, GEO_INTERFACE_ATTR):
        raise TypeError(SHAPE_TYPE_ERR % shape)
    if not hasattr(other, GEO_INTERFACE_ATTR):
        raise TypeError(SHAPE_TYPE_ERR % shape)
    o = geom.shape(shape)
    o2 = geom.shape(other)
    res = o.intersection(o2)
    return asShape(res)


def symmetric_difference(shape, other):
    if not hasattr(shape, GEO_INTERFACE_ATTR):
        raise TypeError(SHAPE_TYPE_ERR % shape)
    if not hasattr(other, GEO_INTERFACE_ATTR):
        raise TypeError(SHAPE_TYPE_ERR % shape)
    o = geom.shape(shape)
    o2 = geom.shape(other)
    res = o.symmetric_difference(o2)
    return asShape(res)


def union(shape, other):
    if not hasattr(shape, GEO_INTERFACE_ATTR):
        raise TypeError(SHAPE_TYPE_ERR % shape)
    if not hasattr(other, GEO_INTERFACE_ATTR):
        raise TypeError(SHAPE_TYPE_ERR % shape)
    o = geom.shape(shape)
    o2 = geom.shape(other)
    res = o.union(o2)
    return asShape(res)


def cascaded_union(shapes):
    o = []
    for shape in shapes:
        if not hasattr(shape, GEO_INTERFACE_ATTR):
            raise TypeError(SHAPE_TYPE_ERR % shape)
        o.append(geom.shape(shape))
    res = shops.unary_union(o)
    return asShape(res)


def unary_union(shapes):
    # seems to be the same as cascade_union except that it handles multipart polygons
    if shapely_version < "1.2.16":
        raise Exception(
            "shapely 1.2.16 or higher needed for unary_union; "
            "upgrade shapely or try cascade_union instead"
        )
    o = []
    for shape in shapes:
        if not hasattr(shape, GEO_INTERFACE_ATTR):
            raise TypeError(SHAPE_TYPE_ERR % shape)
        o.append(geom.shape(shape))
    res = shops.unary_union(o)
    return asShape(res)


# Unary predicates
# ----------------
def has_z(shape):
    if not hasattr(shape, GEO_INTERFACE_ATTR):
        raise TypeError(SHAPE_TYPE_ERR % shape)
    o = geom.shape(shape)
    return o.has_z


def is_empty(shape):
    if not hasattr(shape, GEO_INTERFACE_ATTR):
        raise TypeError(SHAPE_TYPE_ERR % shape)
    o = geom.shape(shape)
    return o.is_empty


def is_ring(shape):
    if not hasattr(shape, GEO_INTERFACE_ATTR):
        raise TypeError(SHAPE_TYPE_ERR % shape)
    o = geom.shape(shape)
    return o.is_ring


def is_simple(shape):
    if not hasattr(shape, GEO_INTERFACE_ATTR):
        raise TypeError(SHAPE_TYPE_ERR % shape)
    o = geom.shape(shape)
    return o.is_simple


def is_valid(shape):
    if not hasattr(shape, GEO_INTERFACE_ATTR):
        raise TypeError(SHAPE_TYPE_ERR % shape)
    o = geom.shape(shape)
    return o.is_valid


# Binary predicates
# -----------------
def relate(shape, other):
    if not hasattr(shape, GEO_INTERFACE_ATTR):
        raise TypeError(SHAPE_TYPE_ERR % shape)
    if not hasattr(other, GEO_INTERFACE_ATTR):
        raise TypeError(SHAPE_TYPE_ERR % shape)
    o = geom.shape(shape)
    o2 = geom.shape(other)
    return o.relate(o2)


def contains(shape, other):
    if not hasattr(shape, GEO_INTERFACE_ATTR):
        raise TypeError(SHAPE_TYPE_ERR % shape)
    if not hasattr(other, GEO_INTERFACE_ATTR):
        raise TypeError(SHAPE_TYPE_ERR % shape)
    o = geom.shape(shape)
    o2 = geom.shape(other)
    return o.contains(o2)


def crosses(shape, other):
    if not hasattr(shape, GEO_INTERFACE_ATTR):
        raise TypeError(SHAPE_TYPE_ERR % shape)
    if not hasattr(other, GEO_INTERFACE_ATTR):
        raise TypeError(SHAPE_TYPE_ERR % shape)
    o = geom.shape(shape)
    o2 = geom.shape(other)
    return o.crosses(o2)


def disjoint(shape, other):
    if not hasattr(shape, GEO_INTERFACE_ATTR):
        raise TypeError(SHAPE_TYPE_ERR % shape)
    if not hasattr(other, GEO_INTERFACE_ATTR):
        raise TypeError(SHAPE_TYPE_ERR % shape)
    o = geom.shape(shape)
    o2 = geom.shape(other)
    return o.disjoint(o2)


def equals(shape, other):
    if not hasattr(shape, GEO_INTERFACE_ATTR):
        raise TypeError(SHAPE_TYPE_ERR % shape)
    if not hasattr(other, GEO_INTERFACE_ATTR):
        raise TypeError(SHAPE_TYPE_ERR % shape)
    o = geom.shape(shape)
    o2 = geom.shape(other)
    return o.equals(o2)


def intersects(shape, other):
    if not hasattr(shape, GEO_INTERFACE_ATTR):
        raise TypeError(SHAPE_TYPE_ERR % shape)
    if not hasattr(other, GEO_INTERFACE_ATTR):
        raise TypeError(SHAPE_TYPE_ERR % shape)
    o = geom.shape(shape)
    o2 = geom.shape(other)
    return o.intersects(o2)


def overlaps(shape, other):
    if not hasattr(shape, GEO_INTERFACE_ATTR):
        raise TypeError(SHAPE_TYPE_ERR % shape)
    if not hasattr(other, GEO_INTERFACE_ATTR):
        raise TypeError(SHAPE_TYPE_ERR % shape)
    o = geom.shape(shape)
    o2 = geom.shape(other)
    return o.overlaps(o2)


def touches(shape, other):
    if not hasattr(shape, GEO_INTERFACE_ATTR):
        raise TypeError(SHAPE_TYPE_ERR % shape)
    if not hasattr(other, GEO_INTERFACE_ATTR):
        raise TypeError(SHAPE_TYPE_ERR % shape)
    o = geom.shape(shape)
    o2 = geom.shape(other)
    return o.touches(o2)


def within(shape, other):
    if not hasattr(shape, GEO_INTERFACE_ATTR):
        raise TypeError(SHAPE_TYPE_ERR % shape)
    if not hasattr(other, GEO_INTERFACE_ATTR):
        raise TypeError(SHAPE_TYPE_ERR % shape)
    o = geom.shape(shape)
    o2 = geom.shape(other)
    return o.within(o2)


def equals_exact(shape, other, tolerance):
    if not hasattr(shape, GEO_INTERFACE_ATTR):
        raise TypeError(SHAPE_TYPE_ERR % shape)
    if not hasattr(other, GEO_INTERFACE_ATTR):
        raise TypeError(SHAPE_TYPE_ERR % shape)
    o = geom.shape(shape)
    o2 = geom.shape(other)
    return o.equals_exact(o2, tolerance)


def almost_equals(shape, other, decimal=6):
    if not hasattr(shape, GEO_INTERFACE_ATTR):
        raise TypeError(SHAPE_TYPE_ERR % shape)
    if not hasattr(other, GEO_INTERFACE_ATTR):
        raise TypeError(SHAPE_TYPE_ERR % shape)
    o = geom.shape(shape)
    o2 = geom.shape(other)
    return o.almost_equals(o2, decimal)


# Linear referencing
# ------------------
def project(shape, other, normalized=False):
    if not hasattr(shape, GEO_INTERFACE_ATTR):
        raise TypeError(SHAPE_TYPE_ERR % shape)
    if not hasattr(other, GEO_INTERFACE_ATTR):
        raise TypeError(SHAPE_TYPE_ERR % shape)
    o = geom.shape(shape)
    o2 = geom.shape(other)
    return o.project(o2, normalized)


def interpolate(shape, distance, normalized=False):
    if not hasattr(shape, GEO_INTERFACE_ATTR):
        raise TypeError(SHAPE_TYPE_ERR % shape)
    o = geom.shape(shape)
    res = o.interpolate(distance, normalized)
    return asShape(res)


# Copy doc strings from shapely
for method in __all__:
    if hasattr(_basegeom, method):
        locals()[method].__doc__ = getattr(_basegeom, method).__doc__



import geopandas as gpd
import pytest
from geodatasets import get_path

from libpysal import graph

pytest.importorskip("h3")
pytest.importorskip("tobler")


class TestH3:
    def setup_method(self):
        from tobler.util import h3fy

        gdf = gpd.read_file(get_path("geoda guerry"))
        h3_geoms = h3fy(gdf, resolution=4)
        self.h3_ids = h3_geoms.index

    def test_h3(self):
        g = graph.Graph.build_h3(self.h3_ids)
        assert g.n == len(self.h3_ids)
        assert g.pct_nonzero == 1.69921875
        assert len(g) == 1740
        assert g.adjacency.max() == 1

    @pytest.mark.parametrize("order", range(2, 6))
    def test_h3_order(self, order):
        g = graph.Graph.build_h3(self.h3_ids, order)
        assert g.n == len(self.h3_ids)
        assert g.adjacency.max() == order

    def test_h3_binary(self):
        g = graph.Graph.build_h3(self.h3_ids, order=4, weight="binary")
        assert g.n == len(self.h3_ids)
        assert g.adjacency.max() == 1



import contextlib


def _read_gal(path):
    """Read GAL weights to Graph object

    Parameters
    ----------
    path : str
        path to GAL file

    Returns
    -------
    dict
        neighbors dict
    """
    with open(path) as file:
        neighbors = {}

        # handle case where more than n is specified in first line
        header = file.readline().strip().split()
        header_n = len(header)
        n = int(header[0])

        if header_n > 1:
            n = int(header[1])

        for _ in range(n):
            id_, _ = file.readline().strip().split()
            neighbors_i = file.readline().strip().split()
            neighbors[id_] = neighbors_i

    # try casting to ints to ensure loss-less roundtrip of integer node ids
    with contextlib.suppress(ValueError):
        neighbors = {int(k): list(map(int, v)) for k, v in neighbors.items()}

    return neighbors


def _to_gal(graph_obj, path):
    """Write GAL weights to Graph object

    Parameters
    ----------
    graph_obj : Graph
        Graph object
    path : str
        path to GAL file
    """
    grouper = graph_obj._adjacency.groupby(level=0, sort=False)

    with open(path, "w") as file:
        file.write(f"{graph_obj.n}\n")

        for ix, chunk in grouper:
            if ix in graph_obj.isolates:
                neighbors = []
            else:
                neighbors = (
                    chunk.index.get_level_values("neighbor").astype(str).tolist()
                )

            file.write(f"{ix} {len(neighbors)}\n")
            file.write(" ".join(neighbors) + "\n")



import os
import tempfile

import pytest

from .... import examples as pysal_examples
from ...fileio import FileIO
from ..gwt import GwtIO


class TesttestGwtIO:
    def setup_method(self):
        self.test_file = test_file = pysal_examples.get_path("juvenile.gwt")
        self.obj = GwtIO(test_file, "r")

    def test_close(self):
        f = self.obj
        f.close()
        pytest.raises(ValueError, f.read)

    def test_read(self):
        w = self.obj.read()
        assert w.n == 168
        assert w.mean_neighbors == 16.678571428571427
        w.transform = "B"
        assert list(w[1].values()) == [1.0]

    def test_seek(self):
        self.test_read()
        pytest.raises(StopIteration, self.obj.read)
        self.obj.seek(0)
        self.test_read()

    # Commented out by CRS, GWT 'w' mode removed until we
    # can find a good solution for retaining distances.
    # see issue #153.
    # Added back by CRS,
    def test_write(self):
        with pytest.warns(RuntimeWarning, match="DBF relating to GWT was not found"):
            w = self.obj.read()
            f = tempfile.NamedTemporaryFile(suffix=".gwt")
            fname = f.name
            f.close()
            o = FileIO(fname, "w")
            # copy the shapefile and ID variable names from the old gwt.
            # this is only available after the read() method has been called.
            # o.shpName = self.obj.shpName
            # o.varName = self.obj.varName
            o.write(w)
            o.close()
            wnew = FileIO(fname, "r").read()
            assert wnew.pct_nonzero == w.pct_nonzero
            os.remove(fname)



#!/usr/bin/env python3
import errno
import os
import platform
import tempfile
from unittest.mock import patch

import pandas
from platformdirs import user_data_dir

from .. import available, get_url, load_example
from ..base import get_data_home

os_name = platform.system()

original_path_exists = os.path.exists
original_makedirs = os.makedirs


class TestExamples:
    def test_available(self):
        examples = available()
        assert type(examples) == pandas.core.frame.DataFrame
        assert examples.shape == (99, 3)

    def test_data_home(self):
        pth = get_data_home()
        head, tail = os.path.split(pth)
        assert tail == "pysal"
        if os_name == "Linux":
            if "XDG_DATA_HOME" in os.environ:
                assert head == os.environ["XDG_DATA_HOME"]
            else:
                heads = head.split("/")
                assert heads[-1] == "share"
                assert heads[-2] == ".local"
        elif os_name == "Darwin":
            heads = head.split("/")
            assert heads[1] == "Users"
            assert heads[-1] == "Application Support"
            assert heads[-2] == "Library"
        elif os_name == "Windows":
            heads = head.split("\\")
            assert heads[1] == "Users"
            assert heads[-2] == "Local"
            assert heads[-3] == "AppData"

    @patch("os.makedirs")
    @patch("os.path.exists")
    def test_data_home_fallback(self, path_exists_mock, makedirs_mock):
        data_home = user_data_dir("pysal", "pysal")

        def makedirs_side_effect(path, exist_ok=False):  # noqa: ARG001
            if path == data_home:
                raise OSError(errno.EROFS)

        def path_exists_side_effect(path):
            if path == data_home:
                return False
            return original_path_exists(path)

        makedirs_mock.side_effect = makedirs_side_effect
        path_exists_mock.side_effect = path_exists_side_effect

        pth = get_data_home()
        head, tail = os.path.split(pth)

        assert tail == "pysal"
        assert head == tempfile.gettempdir()

    def test_get_url(self):
        assert get_url("10740") is None
        url = "https://geodacenter.github.io/data-and-lab//data/baltimore.zip"
        assert get_url("Baltimore") == url

    def test_load_example(self):
        taz = load_example("taz")
        flist = taz.get_file_list()
        assert len(flist) == 4



# ruff: noqa: N999

from ...examples import get_path
from ..fileio import FileIO


def test_by_col_exists():
    """Test if the Metaclass is initializing and providing readers to its children."""

    fh1 = FileIO.open(get_path("columbus.dbf"))
    fh2 = FileIO.open(get_path("usjoin.csv"))

    assert hasattr(fh1, "by_col")
    assert hasattr(fh2, "by_col")



"""

For completeness, we need to test a shuffled dataframe
(i.e. always send unsorted data) with:
- numeric ids
- string ids
- mixed polygon/multipolygon dataset
- mixed line/multiline dataset
- dataset with islands
"""

import geodatasets
import geopandas
import numpy
import pandas
import pytest
import shapely

from libpysal.graph._contiguity import (
    _block_contiguity,
    _fuzzy_contiguity,
    _queen,
    _rook,
    _vertex_set_intersection,
)


@pytest.fixture(scope="session")
def rivers():
    numpy.random.seed(111211)
    rivers = geopandas.read_file(geodatasets.get_path("eea large_rivers")).sample(
        frac=1, replace=False
    )
    rivers["strID"] = rivers.NAME
    rivers["intID"] = rivers.index.values + 2
    return rivers


@pytest.fixture(scope="session")
def nybb():
    nybb = geopandas.read_file(geodatasets.get_path("ny bb"))
    nybb["strID"] = nybb.BoroName
    nybb["intID"] = nybb.BoroCode
    return nybb


parametrize_ids = pytest.mark.parametrize("ids", [None, "strID", "intID"])
parametrize_perim = pytest.mark.parametrize(
    "by_perimeter", [False, True], ids=["binary", "perimeter"]
)
parametrize_rook = pytest.mark.parametrize("rook", [True, False], ids=["rook", "queen"])
parametrize_pointset = pytest.mark.parametrize(
    "pointset", [True, False], ids=["pointset", "vertex intersection"]
)


@pytest.mark.network
@parametrize_pointset
@parametrize_rook
@parametrize_ids
def test_user_rivers(ids, rook, pointset, rivers):
    """
    Check whether contiguity is constructed correctly for rivers in Europe.
    """
    data = rivers.reset_index(drop=False).rename(columns={"index": "original_index"})
    ids = "original_index" if ids is None else ids
    data.index = data[ids].values
    ids = data.index.values
    # implement known_heads, known_tails

    if rook:
        known_heads = known_tails = ids[numpy.arange(len(data))]
        known_weights = numpy.zeros_like(known_heads)
    else:
        known_heads = numpy.array(["Sava", "Danube", "Tisa", "Danube"])
        known_tails = numpy.array(["Danube", "Sava", "Danube", "Tisa"])
        isolates = data[~data.strID.isin(known_heads)].index.values

        tmp_ = (
            data.reset_index(drop=False)
            .rename(columns={"index": "tmp_index"})
            .set_index("strID")
        )

        known_heads = tmp_.loc[known_heads, "tmp_index"].values
        known_tails = tmp_.loc[known_tails, "tmp_index"].values

        known_heads = numpy.hstack((known_heads, isolates))
        known_tails = numpy.hstack((known_tails, isolates))

        known_weights = numpy.ones_like(known_heads)
        known_weights[known_heads == known_tails] = 0
    if pointset:
        f = _rook if rook else _queen
        derived = f(data, ids=ids)
        derived_by_index = f(data, ids=None)
    else:
        derived = _vertex_set_intersection(data, ids=ids, rook=rook)
        derived_by_index = _vertex_set_intersection(data, rook=rook, ids=None)

    assert set(zip(*derived, strict=True)) == set(
        zip(known_heads, known_tails, known_weights, strict=True)
    )
    assert set(zip(*derived_by_index, strict=True)) == set(
        zip(known_heads, known_tails, known_weights, strict=True)
    )


@pytest.mark.network
@parametrize_rook
@parametrize_perim
@parametrize_ids
def test_user_vertex_set_intersection_nybb(ids, rook, by_perimeter, nybb):
    """
    check whether vertexset contiguity is constructed correctly
    for nybb
    """
    data = nybb.copy()
    if ids is not None:
        data.index = data[ids].values
    ids = data.index.values

    # implement known_heads, known_tails
    known_heads = numpy.array([1, 1, 1, 2, 2, 3, 3, 3, 4, 4, 0])
    known_tails = numpy.array([2, 3, 4, 1, 3, 2, 1, 4, 1, 3, 0])

    known_heads = data.index.values[known_heads]
    known_tails = data.index.values[known_tails]

    if by_perimeter:
        head_geom = data.geometry.loc[known_heads].values
        tail_geom = data.geometry.loc[known_tails].values
        known_weights = head_geom.intersection(tail_geom).length
    else:
        known_weights = numpy.ones_like(known_heads)
    known_weights[known_heads == known_tails] = 0

    f = _rook if rook else _queen
    derived = f(data, by_perimeter=by_perimeter, ids=ids)
    derived_by_index = f(data, by_perimeter=by_perimeter, ids=None)

    assert set(zip(*derived, strict=True)) == set(
        zip(known_heads, known_tails, known_weights, strict=True)
    )
    assert set(zip(*derived_by_index, strict=True)) == set(
        zip(known_heads, known_tails, known_weights, strict=True)
    )


@pytest.mark.network
@parametrize_rook
@parametrize_perim
@parametrize_ids
def test_user_pointset_nybb(ids, by_perimeter, rook, nybb):
    """
    check whether pointset weights are constructed correctly
    for nybb
    """
    data = nybb.copy()
    if ids is not None:
        data.index = data[ids].values
    ids = data.index.values

    # implement known_heads, known_tails
    known_heads = numpy.array([1, 1, 1, 2, 2, 3, 3, 3, 4, 4, 0])
    known_tails = numpy.array([2, 3, 4, 1, 3, 2, 1, 4, 1, 3, 0])

    known_heads = data.index.values[known_heads]
    known_tails = data.index.values[known_tails]

    if by_perimeter:
        head_geom = data.geometry.loc[known_heads].values
        tail_geom = data.geometry.loc[known_tails].values
        known_weights = head_geom.intersection(tail_geom).length
    else:
        known_weights = numpy.ones_like(known_heads)
    known_weights[known_heads == known_tails] = 0

    f = _rook if rook else _queen
    derived = f(data, by_perimeter=by_perimeter, ids=ids)
    derived_by_index = f(data, by_perimeter=by_perimeter, ids=None)

    assert set(zip(*derived, strict=True)) == set(
        zip(known_heads, known_tails, known_weights, strict=True)
    )
    assert set(zip(*derived_by_index, strict=True)) == set(
        zip(known_heads, known_tails, known_weights, strict=True)
    )


@parametrize_pointset
def test_correctness_rook_queen_distinct(pointset):
    """
    Check that queen and rook generate different contiguities in the case of a
    shared point but no edge.
    """
    data = geopandas.GeoSeries((shapely.box(0, 0, 1, 1), shapely.box(1, 1, 2, 2)))
    if pointset:
        rook_ = _rook(data.geometry)
        queen_ = _queen(data.geometry)

    else:
        rook_ = _vertex_set_intersection(data.geometry, rook=True)
        queen_ = _vertex_set_intersection(data.geometry, rook=False)

    assert set(zip(*rook_, strict=True)) != set(zip(*queen_, strict=True))


def test_geom_type_raise():
    """
    Check the error for point geoms
    """
    data = geopandas.GeoSeries((shapely.Point(0, 0), shapely.Point(2, 2)))
    with pytest.raises(ValueError, match="This Graph type is only well-defined"):
        _vertex_set_intersection(data)


@pytest.mark.network
def test_overlap_raise(nybb):
    data = nybb.set_index("BoroName").geometry.copy()
    data.iloc[1] = shapely.union(
        data.iloc[1], shapely.Point(1021176.479, 181374.797).buffer(10000)
    )

    with pytest.raises(ValueError, match="Some geometries overlap."):
        _vertex_set_intersection(data, by_perimeter=True)


def test_correctness_vertex_set_contiguity_distinct():
    """
    Check to ensure that vertex set ignores rook/queen neighbors that share
    an edge whose nodes are *not* in the vertex set. The test case is two
    offset squares
    """
    data = geopandas.GeoSeries((shapely.box(0, 0, 1, 1), shapely.box(0.5, 1, 1.5, 2)))

    vs_rook = _vertex_set_intersection(data, rook=True)

    rook = _rook(data)

    assert set(zip(*vs_rook, strict=True)) != set(zip(*rook, strict=True))

    vs_queen = _vertex_set_intersection(data, rook=False)

    queen = _queen(data)

    assert set(zip(*vs_queen, strict=True)) != set(zip(*queen, strict=True))


@pytest.mark.parametrize(
    "regimes",
    [
        ["n", "n", "s", "s", "e", "e", "w", "w", "e", "j"],
        [0, 0, 2, 2, 3, 3, 4, 4, 3, 1],
    ],
)
def test_block_contiguity(regimes):
    neighbors = _block_contiguity(regimes)
    wn = {
        0: [1],
        1: [0],
        2: [3],
        3: [2],
        4: [5, 8],
        5: [4, 8],
        6: [7],
        7: [6],
        8: [4, 5],
        9: [],
    }
    assert {f: n.tolist() for f, n in neighbors.items()} == wn

    ids = ["a", "b", "c", "d", "e", "f", "g", "h", "i", "j"]
    neighbors = _block_contiguity(regimes, ids=ids)
    wn_str = {ids[f]: [ids[o] for o in n] for f, n in wn.items()}
    assert {f: n.tolist() for f, n in neighbors.items()} == wn_str

    regimes = pandas.Series(regimes, index=ids)
    neighbors = _block_contiguity(regimes)
    assert {f: n.tolist() for f, n in neighbors.items()} == wn_str


@pytest.mark.network
def test_fuzzy_contiguity(nybb):
    # integer
    head, tail, weight = _fuzzy_contiguity(nybb.set_index("intID"), nybb["intID"])
    numpy.testing.assert_array_equal(
        head,
        [5, 4, 4, 4, 3, 3, 1, 1, 1, 2, 2],
    )
    numpy.testing.assert_array_equal(
        tail,
        [5, 3, 1, 2, 4, 1, 4, 3, 2, 4, 1],
    )
    numpy.testing.assert_array_equal(weight, [0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1])

    # string
    head, tail, weight = _fuzzy_contiguity(nybb.set_index("strID"), nybb["strID"])
    numpy.testing.assert_array_equal(
        head,
        [
            "Staten Island",
            "Queens",
            "Queens",
            "Queens",
            "Brooklyn",
            "Brooklyn",
            "Manhattan",
            "Manhattan",
            "Manhattan",
            "Bronx",
            "Bronx",
        ],
    )
    numpy.testing.assert_array_equal(
        tail,
        [
            "Staten Island",
            "Brooklyn",
            "Manhattan",
            "Bronx",
            "Queens",
            "Manhattan",
            "Queens",
            "Brooklyn",
            "Bronx",
            "Queens",
            "Manhattan",
        ],
    )
    numpy.testing.assert_array_equal(weight, [0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1])

    # tolerance
    head, tail, weight = _fuzzy_contiguity(
        nybb.set_index("intID"), nybb["intID"], tolerance=0.05
    )
    numpy.testing.assert_array_equal(
        head,
        [5, 4, 4, 4, 3, 3, 3, 1, 1, 1, 2, 2],
    )
    numpy.testing.assert_array_equal(
        tail,
        [3, 3, 1, 2, 5, 4, 1, 4, 3, 2, 4, 1],
    )
    numpy.testing.assert_array_equal(weight, [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1])

    # buffer
    head, tail, weight = _fuzzy_contiguity(
        nybb.set_index("intID"), nybb["intID"], buffer=5000
    )
    numpy.testing.assert_array_equal(
        head,
        [5, 4, 4, 4, 3, 3, 3, 1, 1, 1, 2, 2],
    )
    numpy.testing.assert_array_equal(
        tail,
        [3, 3, 1, 2, 5, 4, 1, 4, 3, 2, 4, 1],
    )
    numpy.testing.assert_array_equal(weight, [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1])

    # predicate
    head, tail, weight = _fuzzy_contiguity(
        nybb.set_index("intID"), nybb["intID"], predicate="within"
    )
    numpy.testing.assert_array_equal(
        head,
        [5, 4, 3, 1, 2],
    )
    numpy.testing.assert_array_equal(
        tail,
        [5, 4, 3, 1, 2],
    )
    numpy.testing.assert_array_equal(weight, [0, 0, 0, 0, 0])

    with pytest.raises(ValueError, match="Only one"):
        _fuzzy_contiguity(
            nybb.set_index("intID"), nybb["intID"], tolerance=0.05, buffer=5000
        )

    # kwargs
    head, tail, weight = _fuzzy_contiguity(
        nybb.set_index("intID"), nybb["intID"], buffer=5000, resolution=2
    )
    numpy.testing.assert_array_equal(
        head,
        [5, 4, 4, 4, 3, 3, 3, 1, 1, 1, 2, 2],
    )
    numpy.testing.assert_array_equal(
        tail,
        [3, 3, 1, 2, 5, 4, 1, 4, 3, 2, 4, 1],
    )
    numpy.testing.assert_array_equal(weight, [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1])



import warnings

import numpy
import pandas
from scipy import sparse
from scipy.spatial import Delaunay as _Delaunay

from libpysal.weights import WSP, W

try:
    from numba import njit
except ModuleNotFoundError:
    from libpysal.common import jit as njit

__author__ = """"
Levi John Wolf (levi.john.wolf@gmail.com)
Martin Fleischmann (martin@martinfleischmann.net)
"""

#### Classes


class Delaunay(W):
    """
    Constructor of the Delaunay graph of a set of input points.
    Relies on scipy.spatial.Delaunay and numba to quickly construct
    a graph from the input set of points. Will be slower without numba,
    and will warn if this is missing.

    Parameters
    ----------
    coordinates :   array of points, (N,2)
        numpy array of coordinates containing locations to compute the
        delaunay triangulation
    **kwargs    :   keyword argument list
        keyword arguments passed directly to weights.W

    Notes
    -----
    The Delaunay triangulation can result in quite a few non-local links among
    spatial coordinates. For a more useful graph, consider the weights.Voronoi
    constructor or the Gabriel graph.

    The weights.Voronoi class builds a voronoi diagram among the points, clips the
    Voronoi cells, and then constructs an adjacency graph among the clipped cells.
    This graph among the clipped Voronoi cells generally represents the structure
    of local adjacencies better than the "raw" Delaunay graph.

    The weights.gabriel.Gabriel graph constructs a Delaunay graph, but only
    includes the "short" links in the Delaunay graph.

    However, if the unresricted Delaunay triangulation is needed, this class
    will compute it much more quickly than Voronoi(coordinates, clip=None).
    """

    def __init__(self, coordinates, **kwargs):
        try:
            from numba import njit  # noqa: F401
        except ModuleNotFoundError:
            warnings.warn(
                "The numba package is used extensively in this module"
                " to accelerate the computation of graphs. Without numba,"
                " these computations may become unduly slow on large data.",
                stacklevel=2,
            )
        edges, _ = self._voronoi_edges(coordinates)
        ids = kwargs.get("ids")
        if ids is not None:
            ids = numpy.asarray(ids)
            edges = numpy.column_stack((ids[edges[:, 0]], ids[edges[:, 1]]))
            del kwargs["ids"]
        else:
            ids = numpy.arange(coordinates.shape[0])

        voronoi_neighbors = pandas.DataFrame(edges).groupby(0)[1].apply(list).to_dict()
        W.__init__(self, voronoi_neighbors, id_order=list(ids), **kwargs)

    def _voronoi_edges(self, coordinates):
        dt = _Delaunay(coordinates)
        edges = _edges_from_simplices(dt.simplices)
        edges = (
            pandas.DataFrame(numpy.asarray(list(edges)))
            .sort_values([0, 1])
            .drop_duplicates()
            .values
        )
        return edges, dt

    @classmethod
    def from_dataframe(cls, df, geom_col=None, ids=None, use_index=None, **kwargs):
        """
        Construct a Delaunay triangulation from a geopandas GeoDataFrame.
        Not that the input geometries in the dataframe must be Points.
        Polygons or lines must be converted to points (e.g. using
        df.geometry.centroid).

        Parameters
        ----------
        df  :   geopandas.GeoDataFrame
            GeoDataFrame containing points to construct the Delaunay
            Triangulation.
        geom_col :  string
            the name of the column in `df` that contains the
            geometries. Defaults to active geometry column.
        ids   :   list-like, string
            a list-like of ids to use to index the spatial weights object or
            the name of the column to use as IDs. If nothing is
            provided, the dataframe index is used if `use_index=True` or
            a positional index is used if `use_index=False`.
            Order of the resulting W is not respected from this list.
        use_index  : bool
            use index of `df` as `ids` to index the spatial weights object.
        **kwargs :  keyword arguments
            Keyword arguments that are passed downwards to the weights.W
            constructor.
        """
        if isinstance(df, pandas.Series):
            df = df.to_frame("geometry")
        if geom_col is None:
            geom_col = df.geometry.name
        geomtypes = df[geom_col].geom_type.unique()

        if ids is None:
            if use_index is None:
                warnings.warn(
                    "`use_index` defaults to False but will default to True in future. "
                    "Set True/False directly to control this behavior and silence this "
                    "warning",
                    FutureWarning,
                    stacklevel=2,
                )
                use_index = False
            if use_index:
                ids = df.index.tolist()

        elif isinstance(ids, str):
            ids = df[ids].tolist()

        try:
            assert len(geomtypes) == 1
            assert geomtypes[0] == "Point"
            point_array = numpy.column_stack(
                (df[geom_col].x.values, df[geom_col].y.values)
            )
            return cls(point_array, ids=ids, **kwargs)
        except AssertionError:
            raise TypeError(
                f"The input dataframe has geometry types {geomtypes}"
                f" but this delaunay triangulation is only well-defined for points."
                f" Choose a method to convert your dataframe into points (like using"
                f" the df.centroid) and use that to estimate this graph."
            ) from None


class Gabriel(Delaunay):
    """
    Constructs the Gabriel graph of a set of points. This graph is a subset of
    the Delaunay triangulation where only "short" links are retained. This
    function is also accelerated using numba, and implemented on top of the
    scipy.spatial.Delaunay class.

    For a link (i,j) connecting node i to j in the Delaunay triangulation
    to be retained in the Gabriel graph, it must pass a point set exclusion test:

    1. Construct the circle C_ij containing link (i,j) as its diameter
    2. If any other node k is contained within C_ij, then remove link (i,j)
       from the graph.
    3. Once all links are evaluated, the remaining graph is the Gabriel graph.

    Parameters
    ----------
    coordinates :   array of points, (N,2)
        numpy array of coordinates containing locations to compute the
        delaunay triangulation
    **kwargs    :   keyword argument list
        keyword arguments passed directly to weights.W
    """

    def __init__(self, coordinates, **kwargs):
        try:
            from numba import njit  # noqa: F401
        except ModuleNotFoundError:
            warnings.warn(
                "The numba package is used extensively in this module"
                " to accelerate the computation of graphs. Without numba,"
                " these computations may become unduly slow on large data.",
                stacklevel=2,
            )
        edges, dt = self._voronoi_edges(coordinates)
        droplist = _filter_gabriel(
            edges,
            dt.points,
        )
        output = numpy.row_stack(list(set(map(tuple, edges)).difference(set(droplist))))
        ids = kwargs.get("ids")
        if ids is not None:
            ids = numpy.asarray(ids)
            output = numpy.column_stack((ids[output[:, 0]], ids[output[:, 1]]))
            del kwargs["ids"]
        else:
            ids = numpy.arange(coordinates.shape[0])

        gabriel_neighbors = pandas.DataFrame(output).groupby(0)[1].apply(list).to_dict()
        W.__init__(self, gabriel_neighbors, id_order=list(ids), **kwargs)


class Relative_Neighborhood(Delaunay):  # noqa: N801
    """
    Constructs the Relative Neighborhood graph from a set of points.
    This graph is a subset of the Delaunay triangulation, where only
    "relative neighbors" are retained. Further, it is a superset of
    the Minimum Spanning Tree, with additional "relative neighbors"
    introduced.

    A relative neighbor pair of points i,j must be closer than the
    maximum distance between i (or j) and each other point k.
    This means that the points are at least as close to one another
    as they are to any other point.

    Parameters
    ----------
    coordinates :   array of points, (N,2)
        numpy array of coordinates containing locations to compute the
        delaunay triangulation
    **kwargs    :   keyword argument list
        keyword arguments passed directly to weights.W
    """

    def __init__(self, coordinates, binary=True, **kwargs):
        try:
            from numba import njit  # noqa: F401
        except ModuleNotFoundError:
            warnings.warn(
                "The numba package is used extensively in this module"
                " to accelerate the computation of graphs. Without numba,"
                " these computations may become unduly slow on large data.",
                stacklevel=2,
            )
        edges, dt = self._voronoi_edges(coordinates)
        output, dkmax = _filter_relativehood(edges, dt.points, return_dkmax=False)
        row, col, data = zip(*output, strict=True)
        if binary:
            data = numpy.ones_like(col, dtype=float)
        sp = sparse.csc_matrix((data, (row, col)))  # TODO: faster way than this?
        ids = kwargs.get("ids")
        if ids is None:
            ids = numpy.arange(sp.shape[0])
        else:
            del kwargs["ids"]
        ids = list(ids)
        tmp = WSP(sp, id_order=ids).to_W()
        W.__init__(self, tmp.neighbors, tmp.weights, id_order=ids, **kwargs)


#### utilities


@njit
def _edges_from_simplices(simplices):
    """
    Construct the sets of links that correspond to the edges of each
    simplex. Each simplex has three "sides," and thus six undirected
    edges. Thus, the input should be a list of three-length tuples,
    that are then converted into the six non-directed edges for
    each simplex.
    """
    edges = []
    for simplex in simplices:
        edges.append((simplex[0], simplex[1]))
        edges.append((simplex[1], simplex[0]))
        edges.append((simplex[1], simplex[2]))
        edges.append((simplex[2], simplex[1]))
        edges.append((simplex[2], simplex[0]))
        edges.append((simplex[0], simplex[2]))
    return numpy.asarray(edges)


@njit
def _filter_gabriel(edges, coordinates):
    """
    For an input set of edges and coordinates, filter the input edges
    depending on the Gabriel rule:

    For each simplex, let i,j be the diameter of the circle defined by
    edge (i,j), and let k be the third point defining the simplex. The
    limiting case for the Gabriel rule is when k is also on the circle
    with diameter (i,j). In this limiting case, then simplex ijk must
    be a right triangle, and dij**2 = djk**2 + dki**2 (by thales theorem).

    This means that when dij**2 > djk**2 + dki**2, then k is inside the circle.
    In contrast, when dij**2 < djk**2 + dji*2, k is outside of the circle.

    Therefore, it's sufficient to take each observation i, iterate over its
    Delaunay neighbors j,k, and remove links whre dij**2 > djk**2 + dki**2
    in order to construct the Gabriel graph.
    """
    edge_pointer = 0
    n_edges = len(edges)
    to_drop = []
    while edge_pointer < n_edges:
        edge = edges[edge_pointer]
        cardinality = 0
        # look ahead to find all neighbors of edge[0]
        for joff in range(edge_pointer, n_edges):
            next_edge = edges[joff]
            if next_edge[0] != edge[0]:
                break
            cardinality += 1
        for ix in range(edge_pointer, edge_pointer + cardinality):
            i, j = edges[ix]  # lookahead ensures that i is always edge[0]
            dij2 = ((coordinates[i] - coordinates[j]) ** 2).sum()
            for ix2 in range(edge_pointer, edge_pointer + cardinality):
                _, k = edges[ix2]
                if j == k:
                    continue
                dik2 = ((coordinates[i] - coordinates[k]) ** 2).sum()
                djk2 = ((coordinates[j] - coordinates[k]) ** 2).sum()

                if dij2 > (dik2 + djk2):
                    to_drop.append((i, j))
                    to_drop.append((j, i))
        edge_pointer += cardinality
    return set(to_drop)


@njit
def _filter_relativehood(edges, coordinates, return_dkmax=False):
    """
    This is a direct implementation of the algorithm from Toussaint (1980), RNG-2

    1. Compute the delaunay
    2. for each edge of the delaunay (i,j), compute
       dkmax = max(d(k,i), d(k,j)) for k in 1..n, k != i, j
    3. for each edge of the delaunay (i,j), prune
       if any dkmax is greater than d(i,j)
    """
    n = edges.max()
    out = []
    r = []
    for edge in edges:
        i, j = edge
        pi = coordinates[i]
        pj = coordinates[j]
        dkmax = 0
        dij = ((pi - pj) ** 2).sum() ** 0.5
        prune = False
        for k in range(n):
            pk = coordinates[k]
            dik = ((pi - pk) ** 2).sum() ** 0.5
            djk = ((pj - pk) ** 2).sum() ** 0.5
            distances = numpy.array([dik, djk, dkmax])
            dkmax = distances.max()
            prune = dkmax < dij
            if (not return_dkmax) & prune:
                break
        if prune:
            continue
        out.append((i, j, dij))
        if return_dkmax:
            r.append(dkmax)

    return out, r



"""Miscellaneous file manipulation utilities."""

import pandas as pd

from ..fileio import FileIO


def check_dups(li):
    """Checks duplicates in a list of ID values. ID values must be read in as a list.
    Author(s) -- Luc Anselin <anselin@uchicago.edu>

    Parameters
    ----------
    li : list
        A collection of ID values.

    Returns
    -------
    dups : list
        The duplicate IDs.
    """

    dups = list({x for x in li if li.count(x) > 1})

    return dups


def dbfdups(dbfpath, idvar):
    """Checks duplicates in a ``.dBase`` file ID variable must be specified correctly.
    Author(s) -- Luc Anselin <anselin@uchicago.edu>

    Parameters
    ----------
    dbfpath : str
        The file path to ``.dBase`` file.
    idvar : str
        The ID variable in ``.dBase`` file.

    Returns
    -------
    dups : list
        The duplicate IDs.
    """
    db = FileIO(dbfpath, "r")
    li = db.by_col(idvar)

    dups = list({x for x in li if li.count(x) > 1})

    return dups


def df2dbf(df, dbf_path, my_specs=None):
    """Convert a ``pandas.DataFrame`` into a ``.dbf``. Author(s) --
    Dani Arribas-Bel <D.Arribas-Bel@liverpool.ac.uk>, Luc Anselin <anselin@uchicago.edu>

    Parameters
    ----------
    df : pandas.DataFrame
        Pandas dataframe object to be entirely written out to a ``.dbf``.
    dbf_path : str
        Path to the output ``.dbf``. It is also returned by the function.
    my_specs : list
        A list with the ``field_specs`` to use for each column.
        Defaults to ``None`` and applies the following scheme.
        * ``int`` -- ``('N', 14, 0)``
          * for all ``int`` types
        * ``float`` -- ``('N', 14, 14)``
          * for all ``float`` types
        * ``str`` -- ``('C', 14, 0)``
          * for ``str``, ``object``, and category
          * for all variants for different type sizes

    Returns
    -------
    dbf_path : str
        Path to the output ``.dbf``

    Notes
    -----
    Use of ``dtypes.name`` may not be fully robust, but preferred
    approach of using ``isinstance`` seems too clumsy.
    """
    if my_specs:
        specs = my_specs
    else:
        # new approach using dtypes.name to avoid numpy name issue in type
        type2spec = {
            "int": ("N", 20, 0),
            "int8": ("N", 20, 0),
            "int16": ("N", 20, 0),
            "int32": ("N", 20, 0),
            "int64": ("N", 20, 0),
            "float": ("N", 36, 15),
            "float32": ("N", 36, 15),
            "float64": ("N", 36, 15),
            "str": ("C", 14, 0),
            "object": ("C", 14, 0),
            "category": ("C", 14, 0),
        }
        types = [df[i].dtypes.name for i in df.columns]
        specs = [type2spec[t] for t in types]

    db = FileIO(dbf_path, "w")
    db.header = list(df.columns)
    db.field_spec = specs

    for _i, row in list(df.T.items()):
        db.write(row)
    db.close()

    return dbf_path


def dbf2df(dbf_path, index=None, cols=False, incl_index=False):
    """Read a ``.dbf`` file as a ``pandas.DataFrame``, optionally
    selecting the index variable and which columns are to be loaded.
    Author(s) -- Dani Arribas-Bel <D.Arribas-Bel@liverpool.ac.uk>

    Parameters
    ----------
    dbf_path : str
        Path to the ``.dbf`` file to be read.
    index : str
        Name of the column to be used as the index of the ``pandas.DataFrame``.
    cols : list
        List with the names of the columns to be read into the ``pandas.DataFrame``.
        Defaults to ``False``, which reads the whole ``.dbf``
    incl_index : bool
        If ``True`` index is included in the ``pandas.DataFrame``
        as a column too. Defaults to ``False``.

    Returns
    -------
    df : pandas.DataFrame
        The resultant ``pandas.DataFrame`` object.
    """

    db = FileIO(dbf_path)

    if cols:
        if incl_index:
            cols.append(index)
        vars_to_read = cols
    else:
        vars_to_read = db.header

    data = {var: db.by_col(var) for var in vars_to_read}

    if index:
        index = db.by_col(index)
        db.close()
        return pd.DataFrame(data, index=index, columns=vars_to_read)
    else:
        db.close()
        return pd.DataFrame(data, columns=vars_to_read)


def dbfjoin(dbf1_path, dbf2_path, out_path, joinkey1, joinkey2):
    """Wrapper function to merge two ``.dbf`` files into a new
    ``.dbf`` file. Uses ``dbf2df`` and ``df2dbf`` to read and
    write the ``.dbf`` files into a ``pandas.DataFrame``. Uses
    all default settings for ``dbf2df`` and ``df2dbf`` (see docs
    for specifics). Author(s) --  Luc Anselin <anselin@uchicago.edu>

    Parameters
    ----------
    dbf1_path : str
        Path to the first (left) ``.dbf`` file.
    dbf2_path : str
        Path to the second (right) ``.dbf`` file.
    out_path : str
        Path to the output ``.dbf`` file (returned by the function).
    joinkey1 : str
        Variable name for the key in the first ``.dbf``.
        Must be specified. Key must take unique values.
    joinkey2 : str
        Variable name for the key in the second ``.dbf``.
        Must be specified. Key must take unique values.

    Returns
    -------
    dp : str
        Path to output file.
    """

    df1 = dbf2df(dbf1_path, index=joinkey1)
    df2 = dbf2df(dbf2_path, index=joinkey2)
    dfbig = pd.merge(df1, df2, left_on=joinkey1, right_on=joinkey2, sort=False)
    dp = df2dbf(dfbig, out_path)

    return dp


def dta2dbf(dta_path, dbf_path):
    """Wrapper function to convert a stata ``.dta`` file into a ``.dbf``
    file.  Uses ``df2dbf`` to write the ``.dbf`` files from a ``pandas.DataFrame``.
    Uses all default settings for ``df2dbf`` (see docs for specifics).
    Author(s) -- Luc Anselin <anselin@uchicago.edu>

    Parameters
    ----------
    dta_path : str
        Path to the Stata ``.dta`` file.
    dbf_path : str
        Path to the output ``.dbf`` file.

    Returns
    -------
    dp : str
        path to output file
    """

    db = pd.read_stata(dta_path)
    dp = df2dbf(db, dbf_path)

    return dp



import os
import tempfile
import warnings

import pytest

# import pysal_examples
from .... import examples as pysal_examples
from ...fileio import FileIO
from ..weight_converter import WeightConverter, weight_convert


@pytest.mark.skip("This function is deprecated.")
class TesttestWeightConverter:
    def setup_method(self):
        test_files = [
            "arcgis_ohio.dbf",
            "arcgis_txt.txt",
            "ohio.swm",
            "wmat.dat",
            "wmat.mtx",
            "sids2.gal",
            "juvenile.gwt",
            "geobugs_scot",
            "stata_full.txt",
            "stata_sparse.txt",
            "spat-sym-us.mat",
            "spat-sym-us.wk1",
        ]
        self.test_files = [pysal_examples.get_path(f) for f in test_files]
        dataformats = [
            "arcgis_dbf",
            "arcgis_text",
            None,
            None,
            None,
            None,
            None,
            "geobugs_text",
            "stata_text",
            "stata_text",
            None,
            None,
        ]
        ns = [88, 3, 88, 49, 49, 100, 168, 56, 56, 56, 46, 46]
        self.dataformats = dict(list(zip(self.test_files, dataformats, strict=True)))
        self.ns = dict(list(zip(self.test_files, ns, strict=True)))
        self.fileformats = [
            ("dbf", "arcgis_dbf"),
            ("txt", "arcgis_text"),
            ("swm", None),
            ("dat", None),
            ("mtx", None),
            ("gal", None),
            ("", "geobugs_text"),
            ("gwt", None),
            ("txt", "stata_text"),
            ("mat", None),
            ("wk1", None),
        ]

    def test__set_w(self):
        for f in self.test_files:
            with warnings.catch_warnings(record=True):
                # note: we are just suppressing the warnings here;
                # individual warnings are tested in their specific readers
                warnings.simplefilter("always")
                wc = WeightConverter(f, dataFormat=self.dataformats[f])
            assert wc.w_set() is True
            assert wc.w.n == self.ns[f]

    def test_write(self):
        for f in self.test_files:
            with warnings.catch_warnings(record=True):
                # note: we are just suppressing the warnings here;
                # individual warnings are tested in their specific readers
                warnings.simplefilter("always")
                wc = WeightConverter(f, dataFormat=self.dataformats[f])

            for ext, dataformat in self.fileformats:
                if f.lower().endswith(ext):
                    continue
                temp_f = tempfile.NamedTemporaryFile(suffix=f".{ext}")
                temp_fname = temp_f.name
                temp_f.close()

                with warnings.catch_warnings(record=True):
                    # note: we are just suppressing the warnings here;
                    # individual warnings are tested in their specific readers
                    warnings.simplefilter("always")
                    if ext == "swm":
                        wc.write(temp_fname, useIdIndex=True)
                    elif dataformat is None:
                        wc.write(temp_fname)
                    elif dataformat in ["arcgis_dbf", "arcgis_text"]:
                        wc.write(temp_fname, dataFormat=dataformat, useIdIndex=True)
                    elif dataformat == "stata_text":
                        wc.write(temp_fname, dataFormat=dataformat, matrix_form=True)
                    else:
                        wc.write(temp_fname, dataFormat=dataformat)

                with warnings.catch_warnings(record=True):
                    # note: we are just suppressing the warnings here;
                    # individual warnings are tested in their specific readers
                    warnings.simplefilter("always")
                    if dataformat is None:
                        wnew = FileIO(temp_fname, "r").read()
                    else:
                        wnew = FileIO(temp_fname, "r", dataformat).read()

                if (
                    ext in ["dbf", "swm", "dat", "wk1", "gwt"]
                    or dataformat == "arcgis_text"
                ):
                    assert wnew.n == wc.w.n - len(wc.w.islands)
                else:
                    assert wnew.n == wc.w.n
                os.remove(temp_fname)

    def test_weight_convert(self):
        for f in self.test_files:
            in_file = f
            in_data_format = self.dataformats[f]
            with warnings.catch_warnings(record=True):
                # note: we are just suppressing the warnings here; individual warnings
                #       are tested in their specific readers
                warnings.simplefilter("always")
                if in_data_format is None:
                    in_file = FileIO(in_file, "r")
                else:
                    in_file = FileIO(in_file, "r", in_data_format)
                wold = in_file.read()
                in_file.close()

            for ext, dataformat in self.fileformats:
                if f.lower().endswith(ext):
                    continue
                temp_f = tempfile.NamedTemporaryFile(suffix=f".{ext}")
                out_file = temp_f.name
                temp_f.close()
                out_data_format, use_id_index, matrix_form = dataformat, False, False
                if ext == "swm" or dataformat in ["arcgis_dbf", "arcgis_text"]:
                    use_id_index = True
                elif dataformat == "stata_text":
                    matrix_form = True

                with warnings.catch_warnings(record=True):
                    # note: we are just suppressing the warnings here;
                    # individual warnings are tested in their specific readers
                    warnings.simplefilter("always")
                    weight_convert(
                        in_file,
                        out_file,
                        in_data_format,
                        out_data_format,
                        use_id_index,
                        matrix_form,
                    )

                with warnings.catch_warnings(record=True):
                    # note: we are just suppressing the warnings here;
                    # individual warnings are tested in their specific readers
                    warnings.simplefilter("always")
                    if dataformat is None:
                        wnew = FileIO(out_file, "r").read()
                    else:
                        wnew = FileIO(out_file, "r", dataformat).read()

                if (
                    ext in ["dbf", "swm", "dat", "wk1", "gwt"]
                    or dataformat == "arcgis_text"
                ):
                    assert wnew.n == wold.n - len(wold.islands)
                else:
                    assert wnew.n == wold.n
                os.remove(out_file)



from warnings import warn

import numpy as np
import pytest

from ....examples import get_path
from ....io.geotable import read_files as rf

# from ... import comparators as comp
# from ... import shapely as she
from ...shapes import Chain
from .. import _shapely as sht


@pytest.mark.skip("Skipping shapely during reorg.")
class TestShapely:
    def setup_method(self):
        self.polygons = rf(get_path("Polygon.shp"))
        self.points = rf(get_path("Point.shp"))
        self.lines = rf(get_path("Line.shp"))
        self.target_poly = self.polygons.geometry[2]
        self.target_point = self.points.geometry[1]
        self.target_line = self.lines.geometry[0]

        self.dframes = [self.polygons, self.points, self.lines]
        self.targets = [self.target_poly, self.target_point, self.target_line]

    def compare(self, func_name, df, **kwargs):
        geom_list = df.geometry.tolist()
        shefunc = she.__dict__[func_name]  # noqa: F821
        shtfunc = sht.__dict__[func_name]  # noqa: F821

        try:
            she_vals = (shefunc(geom, **kwargs) for geom in geom_list)
            sht_vals = shtfunc(df, inplace=False, **kwargs)
            sht_list = sht_vals[f"shape_{func_name}"].tolist()
            for tabular, shapely in zip(sht_list, she_vals, strict=True):
                if comp.is_shape(tabular) and comp.is_shape(shapely):  # noqa: F821
                    comp.equal(tabular, shapely)  # noqa: F821
                else:
                    assert tabular == shapely
        except NotImplementedError as e:
            warn(f"The shapely/PySAL bridge is not implemented: {e}.", stacklevel=2)
            return True

    def test_to_wkb(self):
        for df in self.dframes:
            self.compare("to_wkb", df)

    def test_to_wkt(self):
        for df in self.dframes:
            self.compare("to_wkt", df)

    def test_area(self):
        for df in self.dframes:
            self.compare("area", df)

    def test_distance(self):
        for df in self.dframes:
            for other in self.targets:
                self.compare("distance", df, other=other)

    def test_length(self):
        for df in self.dframes:
            self.compare("length", df)

    def test_boundary(self):
        for df in self.dframes:
            self.compare("boundary", df)

    def test_bounds(self):
        for df in self.dframes:
            self.compare("bounds", df)

    def test_centroid(self):
        for df in self.dframes:
            self.compare("centroid", df)

    def test_representative_point(self):
        for df in self.dframes:
            self.compare("representative_point", df)

    def test_convex_hull(self):
        for df in self.dframes:
            self.compare("convex_hull", df)

    def test_envelope(self):
        for df in self.dframes:
            self.compare("envelope", df)

    def test_buffer(self):
        np.random.seed(555)
        for df in self.dframes:
            self.compare("buffer", df, radius=np.random.randint(10))

    def test_simplify(self):
        tol = 0.001
        for df in self.dframes:
            self.compare("simplify", df, tolerance=tol)

    def test_difference(self):
        for df in self.dframes:
            for target in self.targets:
                self.compare("difference", df, other=target)

    def test_intersection(self):
        for df in self.dframes:
            for target in self.targets:
                self.compare("intersection", df, other=target)

    def test_symmetric_difference(self):
        for df in self.dframes:
            for target in self.targets:
                self.compare("symmetric_difference", df, other=target)

    def test_union(self):
        for df in self.dframes:
            for target in self.targets:
                self.compare("union", df, other=target)

    def test_has_z(self):
        for df in self.dframes:
            self.compare("has_z", df)

    def test_is_empty(self):
        """
        PySAL doesn't really support empty shapes. Like, the following errors out:

        ```
        ps.cg.Polygon([[]])
        ```

        and you can make it work by:

        ```
        ps.cg.Polygon([[()]])
        ```

        but that won't convert over to shapely.

        So, we're only testing the negative here.

        """
        for df in self.dframes:
            self.compare("is_empty", df)

    def test_is_ring(self):
        for df in self.dframes:
            self.compare("is_ring", df)

    def test_is_simple(self):
        for df in self.dframes:
            self.compare("is_simple", df)

    def test_is_valid(self):
        for df in self.dframes:
            self.compare("is_valid", df)

    def test_relate(self):
        for df in self.dframes:
            for target in self.targets:
                self.compare("relate", df, other=target)

    def test_contains(self):
        for df in self.dframes:
            for target in self.targets:
                self.compare("contains", df, other=target)

    def test_crosses(self):
        for df in self.dframes:
            for target in self.targets:
                self.compare("crosses", df, other=target)

    def test_disjoint(self):
        for df in self.dframes:
            for target in self.targets:
                self.compare("disjoint", df, other=target)

    def test_equals(self):
        for df in self.dframes:
            for target in self.targets:
                self.compare("equals", df, other=target)

    def test_intersects(self):
        for df in self.dframes:
            for target in self.targets:
                self.compare("intersects", df, other=target)

    def test_overlaps(self):
        for df in self.dframes:
            for target in self.targets:
                self.compare("overlaps", df, other=target)

    def test_touches(self):
        for df in self.dframes:
            for target in self.targets:
                self.compare("touches", df, other=target)

    def test_within(self):
        for df in self.dframes:
            for target in self.targets:
                self.compare("within", df, other=target)

    def test_equals_exact(self):
        for df in self.dframes:
            for target in self.targets:
                self.compare("equals_exact", df, other=target, tolerance=0.1)

    def test_almost_equals(self):
        for df in self.dframes:
            for target in self.targets:
                self.compare("almost_equals", df, other=target)

    def test_project(self):
        np.random.seed(555)
        self.compare("project", self.lines, other=self.targets[2])

    def test_interpolate(self):
        np.random.seed(555)
        for df in self.dframes:
            if isinstance(df.geometry[0], Chain):
                self.compare("interpolate", df, distance=np.random.randint(10))
            else:
                with pytest.raises(TypeError):
                    self.compare("interpolate", df, distance=np.random.randint(10))
